 /*****************************************************************************
 *                                                                            *
 *   Copyright (C) 2004-2006 by Michael Schulze                               *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include "ipod.h"

using namespace itunesdb;

IPod::IPod( const QString& mountPoint )
    :   m_ipodBase( mountPoint ), m_sysInfo( mountPoint ), m_itunesdb( m_sysInfo ),
        m_deviceDetails( mountPoint + "/" + IPOD_CONTROLDIR + "/iTunes/DeviceInfo" )
{
}


IPod::~IPod()
{
    m_itunesdb.clear();
}

bool IPod::open() {
    if ( isOpen() ) {
        return true;
    }

    if ( !m_sysInfo.load() ) {
        return false;
    }

    if ( !m_itunesdb.open() ) {
        return false;
    }

    if ( !m_deviceDetails.load() || m_deviceDetails.getName().isEmpty() ) {
        // get devicename from the itunesdb and set it here to make sure we don't get garbage here
        m_deviceDetails.setName( m_itunesdb.getTitle() );
    }

    return true;
}


void IPod::initialize( const QString& title ) {
	// FIXME make this bool so it's possible to check for success
    if ( isOpen() ) {
        return;
    }

    QString ipodCtrlDirName = m_sysInfo.getIPodControlDirName();
    QString iTunesDir( "iTunes" );

    // create iPod_Control Dir
    QDir dir;
    if ( !dir.mkdir( ipodCtrlDirName ) ) {
    	return;
    }
    dir.cd( ipodCtrlDirName );
    dir.mkdir( iTunesDir );
    dir.cd( iTunesDir );

    // initialize Music Dirs
    m_sysInfo.createMusicDirs();
    m_sysInfo.load();

    // initialize database
    QString dbFileName( "iTunesDB" );
    if ( !dir.exists( dbFileName ) ) {
        // write an empty database with the given title
        m_itunesdb.setTitle( title );
        m_itunesdb.writeDatabase( dir.filePath( dbFileName ) );
    }
    m_itunesdb.open();
    m_itunesdb.setTitle( title );

}


bool IPod::isOpen() {
    return m_itunesdb.isOpen();
}

bool IPod::isStillConnected() {
    return ! m_itunesdb.dbFileChanged();
}

void IPod::close() {
    m_itunesdb.clear();
    m_sysInfo.clear();
    m_deviceDetails.clear();
}



void IPod::synchronize() {
    lock(true);
    m_itunesdb.writeDatabase();
    m_deviceDetails.save();
    unlock();
}


const QString& IPod::getItunesDBError() const {
    return m_itunesdb.error;
}


QString IPod::getName() const {
    return m_deviceDetails.getName();
}


void IPod::setName(const QString& name) {
    m_itunesdb.setTitle( name );
    m_deviceDetails.setName( name );
}


IPodSysInfo& IPod::getSysInfo() {
    return m_sysInfo;
}

const IPodSysInfo& IPod::getSysInfo() const {
    return m_sysInfo;
}

ITunesDB& IPod::getITunesDB() {
    return m_itunesdb;
}

const ITunesDB& IPod::getITunesDB() const {
    return m_itunesdb;
}

void IPod::lock(bool write_lock) {
    m_itunesdb.lock(write_lock);
}

bool IPod::isLocked() {
    return m_itunesdb.isLocked();
}


void IPod::unlock() {
    m_itunesdb.unlock();
}


QString IPod::createDistinctIPodName( const IPod& ipod ) {
    QString basePath = ipod.getBasePath();
    return basePath.replace("/", ".").remove(0, 1);
}


QStringList& IPod::findUnreferencedMusicFiles( QStringList& list ) const {

    QDir musicDir( m_sysInfo.getMusicDir() );
    QString pathInfoBase = ":" IPOD_CONTROLDIR ":" + musicDir.dirName() + ":";
    QStringList fDirs = musicDir.entryList( "F??;f??", QDir::Dirs );
    QStringList::iterator dirIter = fDirs.begin();
    for ( ; dirIter != fDirs.end(); ++dirIter ) {
        if ( ! musicDir.cd( *dirIter ) ) {
            continue;
        }
        QStringList files = musicDir.entryList( "*.mp3", QDir::Files );
        QStringList::iterator fileIter = files.begin();
        for ( ; fileIter != files.end(); ++fileIter ) {
            QString filename = *fileIter;
            QString pathInfo = pathInfoBase + musicDir.dirName() + ":" + filename;

            TrackPredicates::ByPathInfo predicate = TrackPredicates::ByPathInfo( pathInfo );
            if ( m_itunesdb.findFirstTrackBy( predicate ) == NULL ) {
                list.append( pathInfo );
            }
        }
        musicDir.cdUp();
    }

    return list;
}

