 /*****************************************************************************
 *   Copyright (C) 2004 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include <sys/stat.h>
#include <sys/file.h>

#include "itunesdb.h"
#include "utils.h"

#include <qfile.h>
#include <qdir.h>
#include <qfileinfo.h>


using namespace itunesdb;


/***
 *
 * ITunesDBPlaylist Impl
 *
 ***/


ITunesDBPlaylist::ITunesDBPlaylist( const ITunesDBPlaylist& source )
    : Playlist( source ), m_itunesdb( source.m_itunesdb )
{
    ConstIterator itemIter = source.getElements();
    while ( itemIter.hasNext() ) {
        const PlaylistItem * item = itemIter.next();
        addPlaylistItem( createNewItem( item->getID() ) );
    }
    if ( source.mSplRuleSet ) {
        mSplRuleSet = new ITunesDBSPLRuleSet( *source.mSplRuleSet, *this, m_itunesdb );
    }
}


ITunesDBPlaylist::ITunesDBPlaylist( ITunesDB& itunesdb )
    : Playlist(), m_itunesdb( itunesdb ) {}

ITunesDBPlaylist::ITunesDBPlaylist( ITunesDB& itunesdb, const Playlist& playlist )
    : Playlist( playlist ), m_itunesdb( itunesdb ) {}


ITunesDBPlaylist::~ITunesDBPlaylist() {
    // try to remove it from the database
    ITunesDB::PlaylistContainer::iterator pos =
        std::find( m_itunesdb.m_playlists->begin(), m_itunesdb.m_playlists->end(), this );

    if ( pos != m_itunesdb.m_playlists->end() ) {
        m_itunesdb.m_playlists->erase( pos );

    }
}


itunesdb::TrackComparatorPtr ITunesDBPlaylist::getTrackComparatorFor( Q_UINT32 sortOrder ) {
    switch ( sortOrder ) {
    case SORTORDER_TITLE:
        return TrackComparators::BY_TITLE;
    case SORTORDER_ARTIST:
        return TrackComparators::BY_ARTIST;
    case SORTORDER_ALBUM:
        return TrackComparators::BY_ALBUM;
    case SORTORDER_BITRATE:
        return TrackComparators::BY_BITRATE;
    case SORTORDER_GENRE:
        return TrackComparators::BY_GENRE;
    case SORTORDER_FILETYPE:
        return TrackComparators::BY_FILETYPE;
    case SORTORDER_TIME_MODIFIED:
        return TrackComparators::BY_TIMEMODIFIED;
    case SORTORDER_TRACK_NR:
        return TrackComparators::BY_TRACKNUMBER;
    case SORTORDER_SIZE:
        return TrackComparators::BY_SIZE;
    case SORTORDER_TIME:
        return TrackComparators::BY_TRACKLENGTH;
    case SORTORDER_YEAR:
        return TrackComparators::BY_YEAR;
    case SORTORDER_SAMPLERATE:
        return TrackComparators::BY_SAMPLERATE;
    case SORTORDER_COMMENT:
        return TrackComparators::BY_COMMENT;
    case SORTORDER_TIME_ADDED:
        return TrackComparators::BY_DATEADDED;
    case SORTORDER_EQUALIZER:
        return TrackComparators::BY_EQSETTING;
    case SORTORDER_COMPOSER:
        return TrackComparators::BY_COMPOSER;
    case SORTORDER_PLAYCOUNT:
        return TrackComparators::BY_PLAYCOUNT;
    case SORTORDER_TIME_PLAYED:
        return TrackComparators::BY_LASTPLAYED_REVERSE;
    case SORTORDER_CD_NR:
        return TrackComparators::BY_CDNUMBER;
    case SORTORDER_RATING:
        return TrackComparators::BY_RATING;
    case SORTORDER_RELEASE_DATE:
        return TrackComparators::BY_RELEASEDATE;
    case SORTORDER_CATEGORY:
        return TrackComparators::BY_CATEGORY;
    case SORTORDER_DESCRIPTION:
        return TrackComparators::BY_DESCRIPTIONTXT;
    case SORTORDER_BPM:
        return TrackComparators::BY_BPM;
    case SORTORDER_GROUPING:
        return TrackComparators::BY_GROUPING;
    case SORTORDER_MANUAL:
    default:
        return TrackComparatorPtr();
    }
}


PlaylistItem * ITunesDBPlaylist::createNewItem( Q_UINT32 itemid ) {
    int trackpos = m_itunesdb.getMHITTrackPosFor( itemid );
    return new ITunesDBPlaylist::Entry( itemid, trackpos );
}


class TrackBasedComparator : public PlaylistItemComparator {
    TrackComparatorPtr m_trackComparator;
    const ITunesDB& m_itunesdb;
public:
    TrackBasedComparator( const TrackComparatorPtr& trackComparator, const ITunesDB& itunesdb )
        : m_trackComparator( trackComparator ), m_itunesdb( itunesdb ) {}
    virtual ~TrackBasedComparator() {
    }
    virtual int compare( const PlaylistItem& item1, const PlaylistItem& item2 ) const {
        Track * t1 = m_itunesdb.getTrackAt( ((ITunesDBPlaylist::Entry&)item1).getMHITTrackPos() );
        Track * t2 = m_itunesdb.getTrackAt( ((ITunesDBPlaylist::Entry&)item2).getMHITTrackPos() );
        if ( !t1 || !t2 ) {
            return 0;
        }
        return m_trackComparator->compare( *t1, *t2 );
    }
};


void ITunesDBPlaylist::setSortOrder( Playlist::Sortorder sortOrder ) {
    if ( sortOrder == SORTORDER_RANDOM ) {
        m_tracklist.setComparator( NULL );
        m_order = SORTORDER_RANDOM;

    } else {
        // get the comparator for the given sort criteria
        TrackComparatorPtr trackComparator = getTrackComparatorFor( sortOrder );
        if ( !trackComparator ) {
            m_tracklist.setComparator( NULL );
            m_order = SORTORDER_MANUAL;
        } else {
            m_tracklist.setComparator( new TrackBasedComparator( trackComparator, m_itunesdb ) );
            m_order = sortOrder;
        }

    }

    sort();
}


void ITunesDBPlaylist::readNonStringMHOD( QDataStream& stream, Q_UINT32 type, Q_UINT32 blocklen ) {
    switch( type ) {
    case MHOD_LIBRARY_PLAYLIST_INDEX:
        itunesdb::utils::seekRelative( stream, blocklen - 16 );  // skip the whole block
        break;    // ignore
    default:
        Playlist::readNonStringMHOD( stream, type, blocklen );
    }
}


void ITunesDBPlaylist::setDirty( bool dirtyFlag ) {
    Playlist::setDirty( dirtyFlag );
    if ( dirtyFlag ) {
        ITunesDB * nonconst = const_cast< ITunesDB * >( &m_itunesdb );
        nonconst->setDirty();
    }
}


SmartPlaylistRuleSet * ITunesDBPlaylist::createNewSplRuleSet() {
    return new ITunesDBSPLRuleSet( *this, m_itunesdb );
}


ITunesDBPlaylist::Entry::Entry( Q_UINT32 id, uint trackpos )
    : itunesdb::PlaylistItem( id ), m_mhitTrackPos( trackpos )
{
}


ITunesDBPlaylist::Entry::~Entry() {
}


void ITunesDBPlaylist::doPostParsingChecks() {
    Iterator track_iter= getElements();
    while( track_iter.hasNext()) {
        ITunesDBPlaylist::Entry * entry = dynamic_cast<ITunesDBPlaylist::Entry *>( track_iter.next() );
        if ( !entry ) {
            removeTrackAt( track_iter);
            continue;
        }

        int trackpos = m_itunesdb.getMHITTrackPosFor( entry->getID() );
        if ( trackpos < 0 ) {    // track couldn't be found
            removeTrackAt( track_iter);
            continue;
        }

        entry->m_mhitTrackPos = trackpos;
    }
    setSortOrder( getSortOrder() );
}


void ITunesDBPlaylist::updateSmartPlaylist( bool force ) {
    if ( isSmartPlaylist() ) {
        ITunesDBSPLRuleSet * splrules = dynamic_cast<ITunesDBSPLRuleSet *>( mSplRuleSet );
        if ( splrules ) {
            splrules->refreshPlaylist( force );
        }
    }
    sort();
}


ITunesDBTrack * ITunesDBPlaylist::getTrackAt( uint position ) const {
    if ( position >= m_tracklist.count() ) {
        return NULL;
    }
    PlaylistItem * playlistItem = m_tracklist[ position ];
    if ( !playlistItem ) {
        return NULL;
    }
    return getTrack( *playlistItem );
}


ITunesDBTrack * ITunesDBPlaylist::getTrack( const PlaylistItem& playlistItem ) const {
    const ITunesDBPlaylist::Entry * entry = dynamic_cast<const ITunesDBPlaylist::Entry*>( &playlistItem );
    if ( !entry ) {
        // fall back - look it up by trackID
        return m_itunesdb.getTrackByID( playlistItem.getID() );
    }
    ITunesDBTrack * result = m_itunesdb.getTrackAt( entry->m_mhitTrackPos );
    if ( !result ) {
        uint trackpos = m_itunesdb.getMHITTrackPosFor( entry->getID() );
        result = m_itunesdb.getTrackAt( trackpos );
        if ( result ) {
            // dont care about constness since this field is just for caching purposes
            const_cast<ITunesDBPlaylist::Entry *>( entry )->m_mhitTrackPos = trackpos;
        }
    }
    return result;
}


/***
 *
 * ITunesDBTrack impl details
 *
 ***/

ITunesDBTrack::ITunesDBTrack( Q_UINT32 trackid, ITunesDB& itunesdb )
    : Track( trackid ), m_itunesdb( itunesdb )
{
}


ITunesDBTrack::~ITunesDBTrack() {
    // Try to remove from itunesdb without deleting it
    m_itunesdb.removeTrack( getID(), false );
}


void ITunesDBTrack::setDateAdded( Q_UINT32 date ) {
    date_added = date;
}


QString ITunesDBTrack::getFilePath() const {
    return m_itunesdb.getFileForPathInfo( getPathInfo() );
}


void ITunesDBTrack::setDirty( bool dirtyFlag ) {
    itunesdb::Track::setDirty( dirtyFlag );
    if ( dirtyFlag ) {
        ITunesDB * nonconst = const_cast< ITunesDB * >( &m_itunesdb );
        nonconst->setDirty();
    }
}



/***
   *
   *  Predicates
   *
   ***/


struct PlaylistByTitleFinder {
public:
    PlaylistByTitleFinder(const QString& title) : mTitle( title ) { }
    bool operator() ( const Playlist * playlist ) const {
        return playlist->getTitle() == mTitle;
    }
private:
    const QString mTitle;
};


struct PlaylistByIDFinder {
public:
    PlaylistByIDFinder( Q_UINT64 id ) : mID( id ) { }
    bool operator() ( const Playlist * playlist ) const {
        return playlist->getID() == mID;
    }
private:
    const Q_UINT64 mID;
};


struct TrackByDBIDFinder : public TrackPredicate {
    TrackByDBIDFinder( Q_UINT64 dbid ) : mDBID( dbid ) {}
    virtual ~TrackByDBIDFinder() {}
    bool operator () ( const Track * track ) const {
        return track && track->getDBID() == mDBID;
    }
    const Q_UINT64 mDBID;
};


/***
 *
 * ITUNESDB implementation
 *
 ***/

/**
 * An Iterator over a set of tracks
 */
typedef ITunesDB::TrackList::FilteredConstIterator<TrackPredicates::PredicatePtrDelegator> ItDBTrackConstIter;
typedef ITunesDB::TrackList::FilteredIterator<TrackPredicates::PredicatePtrDelegator> ItDBTrackIter;


ITunesDB::ITunesDB( IPodSysInfo& ipodSysInfo ) :
    mInternalVersion( 0 ),
    m_playlists( new PlaylistContainer() ),
    maxtrackid( 0x7E ),
    mMaxDBID( 0 ),
    m_ipodSysInfo( ipodSysInfo ),
    hasPodcastsFlag( false ),
    locked( false )
{
}


ITunesDB::~ITunesDB()
{
    clear();
    delete m_playlists;
}


bool ITunesDB::isOpen() {
    return timestamp.isValid();
}


bool ITunesDB::dbFileChanged() const {
    return !itunesdbfile.exists() || QFileInfo(itunesdbfile.name()).lastModified() != timestamp;
}


/*************************************************
 *
 *       Service Methods
 *
 *************************************************/


ITunesDBPlaylist * ITunesDB::createNewPlaylist( const QString& title ) {
    ITunesDBPlaylist * result = new ITunesDBPlaylist( *this );
    if ( mMaxDBID != 0 ) {  // iTunesDB is open
        result->setID( getNextFreeDBID() );
    }
    if ( !title.isEmpty() ) {
        result->setTitle( title );
    }
    return result;
}


ITunesDBTrack * ITunesDB::createEmptyTrack( Q_UINT32 trackid ) {
    return new ITunesDBTrack( trackid, *this );
}


ITunesDBTrack * ITunesDB::createNewTrack( const QString& fileExtension ) {
    Q_UINT32 trackid = getNextFreeTrackID();
#if 0
	printf( "next free trackID is %d\n", trackid );
#endif
    ITunesDBTrack * track = new ITunesDBTrack( trackid, *this );

    // calculate the directory
    QString trackpath = m_ipodSysInfo.findNewTrackFileName( trackid );

    if ( fileExtension.isEmpty() || !fileExtension.startsWith( "." ) ) {
        trackpath.append( '.' );
    }
    if ( !fileExtension.isEmpty() ) {
        trackpath.append( fileExtension );
    }

    track->setPathInfo( trackpath );
    track->setDBID( getNextFreeDBID() );

    return track;
}

bool compatibleWith( QFile& oldItunesDBFile, QString& error );

bool ITunesDB::addTrack( ITunesDBTrack * track ) {
    if ( track->getID() == 0 ) {
        // not initialized - don't care about this one
        return false;
    }
    if ( getTrackByID( track->getID() ) ) {
        // track already exists
        return false;
    }
    if ( track->getPathInfo().isEmpty() ) {
        // trackpath not set
        return false;
    }
    
    QString lerror;
    if ( !compatibleWith( itunesdbfile, lerror ) ) {
    	return false;
    }

    if ( !insertTrackToDataBase( track ) ) {
        return false;
    }

    track->setDateAdded( QDateTime::currentDateTime().toTime_t() );

    TrackPtrList album;
    getAlbum( track->getArtist(), track->getAlbum(), album );
    uint numTracksInAlbum = album.count();
    TrackPtrList::Iterator albumIter = album.iterator();
    while( albumIter.hasNext() ) {
        albumIter.next()->setTrackCount( numTracksInAlbum );
    }

    // if tracknum is not set yet - set it to the position in the album
    if ( track->getTrackNumber() == 0 ) {
        track->setTrackNumber( numTracksInAlbum );
    }

    track->setDirty();

    return true;
}

bool ITunesDB::addPlaylist( ITunesDBPlaylist * playlist ) {
    if ( getPlaylistByTitle( playlist->getTitle() ) ) {
        // a playlist with the same title already exists
        // printf( "a playlist %s already exists\n", playlist->getTitle().ascii() );
        return false;
    }

    // printf( "adding playlist %s\n", playlist->getTitle().ascii() );

    // check if the DBID is available and itunesdb is open
    if ( isOpen() && getPlaylistByID( playlist->getID() ) ) {
        playlist->setID( getNextFreeDBID() );
    }

    if ( mMaxDBID < playlist->getID() ) {
        mMaxDBID = playlist->getID();
    }

    m_playlists->push_back( playlist );
    setDirty();

    return true;
}


/**
 * returns the Track corresponding to the given ID
 * @param id ID of the track
 * @return the Track corresponding to the given ID
 */
ITunesDBTrack * ITunesDB::getTrackByID( const Q_UINT32 id) const
{
    int trackIndex = getMHITTrackPosFor( id );
    if ( trackIndex < 0 ) {
        return NULL;
    } else {
        return getTrackAt( (uint)trackIndex );
    }
}


ITunesDBTrack * ITunesDB::findTrackByDBID( const Q_UINT64 dbid ) const {
    return findFirstTrackBy( TrackByDBIDFinder( dbid ) );
}


int ITunesDB::getMHITTrackPosFor( Q_UINT32 id ) const {
    TrackIDMap::const_iterator trackIdx = m_TrackIdMap.find( id );
    if ( trackIdx == m_TrackIdMap.end() ) {
        return -1;
    } else {
        return trackIdx->second;
    }
}


ITunesDBTrack * ITunesDB::getTrackAt( uint mhitTrackPos ) const {
    // pre access checks
    if ( mhitTrackPos >= m_TrackList.count() ) {
        return NULL;
    }

    return m_TrackList[ mhitTrackPos ];
}



ITunesDBTrack * ITunesDB::findTrack(const QString& artistname, const QString& albumname, const QString& title, Q_UINT32 trackNum ) const {
    TrackPredicates::ByFullInfo trackByFullNameFinder( artistname, albumname, title, trackNum );
    return findFirstTrackBy( trackByFullNameFinder );
}

/*!
    \fn ITunesDB::getArtists( QStringList &buffer)
 */
QStringList* ITunesDB::getArtists( QStringList &buffer ) const
{
    TrackList::ConstIterator trackIter = m_TrackList.iterator();
    while( trackIter.hasNext() ) {
        const ITunesDBTrack * track = trackIter.next();
        if ( !track ) {
            continue;
        }
        const QString& artist = track->getArtist();
        if ( !buffer.contains( artist ) ) {
            buffer.append( artist );
        }
    }
    return &buffer;
}

QStringList * ITunesDB::getAllAlbums( QStringList& buffer ) const {
    TrackList::ConstIterator trackIter = m_TrackList.iterator();
    while( trackIter.hasNext() ) {
        const ITunesDBTrack * track = trackIter.next();
        if ( !track ) {
            continue;
        }
        const QString& album = track->getAlbum();
        if ( !buffer.contains( album ) ) {
            buffer.append( album );
        }
    }
    return &buffer;
}

/*!
    \fn ITunesDB::getAlbumsByArtist( QStringList &buffer)
 */
QStringList * ITunesDB::getAlbumsByArtist( const QString& artistname, QStringList& buffer ) const {
    TrackPredicates::ByArtist byArtist( artistname );
    ItDBTrackConstIter artistIter = m_TrackList.filteredElements( TrackPredicates::PredicatePtrDelegator( &byArtist ) );
    if ( !artistIter.hasNext() ) {
        return NULL;    // no such artist found
    }
    do {    // hasNext() already checked with the if statement above
        const itunesdb::Track * track = artistIter.next();
        if ( !track ) {
            continue;
        }
        if ( !buffer.contains( track->getAlbum() ) ) {
            buffer.append( track->getAlbum() );
        }
    } while( artistIter.hasNext() );
    return &buffer;
}


/*!
    \fn ITunesDB::containsArtist()
 */
bool ITunesDB::containsArtist( const QString& artistname ) const {
    TrackPredicates::ByArtist tracksByArtistFinder( artistname );
    return m_TrackList.filteredElements( TrackPredicates::PredicatePtrDelegator(  &tracksByArtistFinder ) ).hasNext();
}


TrackPtrList *
ITunesDB::getTracksByArtist( const QString& artistname, TrackPtrList& buffer ) const {
    TrackPredicates::ByArtist tracksByArtistFinder( artistname );
    return getTracksBy( tracksByArtistFinder, buffer );
}


TrackPtrList * ITunesDB::getTracksContaining( const QString& someString, TrackPtrList& buffer, bool caseSensitive ) const {
    return getTracksBy( TrackPredicates::Contains( someString, caseSensitive ), buffer );
}


bool ITunesDB::removeArtist(const QString& artistname) {
    TrackPredicates::ByArtist byArtist( artistname );
    ItDBTrackIter artistIter = m_TrackList.filteredElements( TrackPredicates::PredicatePtrDelegator( &byArtist ) );
    if ( !artistIter.hasNext() ) {
        // no such artist found
        return false;
    }
    // iterators won't get invalidated since we just set the Track to NULL instead of removing it
    while ( artistIter.hasNext() ) {
        removeTrack( artistIter.next()->getID() );
    }
    setDirty();
    return true;
}


QStringList * ITunesDB::getPlaylistTitles( QStringList& buffer ) const {
    PlaylistContainer::iterator playlist = m_playlists->begin();
    for ( ; playlist != m_playlists->end(); ++playlist ) {
        buffer.append( (*playlist)->getTitle() );
    }
    return &buffer;
}


/*!
\fn ITunesDB::getPlaylistByTitle( const QString& playlisttitle)
 */
ITunesDBPlaylist * ITunesDB::getPlaylistByTitle( const QString& playlisttitle) const
{
    if( m_playlists->size() == 0 ) {
        return NULL;    // no playlists
    }
    ITunesDB::PlaylistContainer::iterator pos =
        std::find_if( m_playlists->begin(), m_playlists->end(), PlaylistByTitleFinder( playlisttitle ) );
    if ( pos != m_playlists->end() ) {
        return *pos;
    } else {
        return NULL;
    }
}


ITunesDBPlaylist * ITunesDB::getPlaylistByID( Q_UINT64 playlistID ) const {
    PlaylistByIDFinder finder( playlistID );
    ITunesDB::PlaylistContainer::iterator pos =
        utils::findFirst( m_playlists->begin(), m_playlists->end(), finder );
    if ( pos != m_playlists->end() ) {
        return *pos;
    } else {
        return NULL;
    }
}


/*!
    \fn ITunesDB::getAlbum(QString &artist, QString &album)
 */
TrackPtrList *
ITunesDB::getAlbum(const QString &artistname, const QString &albumname, TrackPtrList& buffer) const {
    TrackPredicates::ByAlbum albumFinder(artistname, albumname);
    return getTracksBy( albumFinder, buffer );
}


bool ITunesDB::containsAlbum( const QString& artistname, const QString& albumname ) const {
    TrackPredicates::ByAlbum albumFinder(artistname, albumname);

    return m_TrackList.filteredElements(
        TrackPredicates::PredicatePtrDelegator( &albumFinder ) ).hasNext();
}


ITunesDBTrack * ITunesDB::findFirstTrackBy( const TrackPredicate& predicate ) const {
    ItDBTrackConstIter trackIter = m_TrackList.filteredElements(
                TrackPredicates::PredicatePtrDelegator( &predicate ) );
    if ( trackIter.hasNext() ) {
        return trackIter.next();
    } else {
        return NULL;
    }
}


TrackPtrList * ITunesDB::getAllTracks( TrackPtrList& buffer ) const {
    buffer.addAll( getAllTracks() );
    return &buffer;
}


TrackPtrList * ITunesDB::getTracksBy( const TrackPredicate& predicate, TrackPtrList& buffer ) const {
    ItDBTrackConstIter trackIter = m_TrackList.filteredElements(
                TrackPredicates::PredicatePtrDelegator( &predicate ) );
    if ( ! trackIter.hasNext() ) {
        return NULL;
    }
    do {
        buffer.append( trackIter.next() );
    } while ( trackIter.hasNext() );


    return &buffer;
}


/*!
    \fn ITunesDB::clear()
 */
void ITunesDB::clear()
{
    // clear playlists
    PlaylistContainer::iterator iter = m_playlists->begin();
    while( iter != m_playlists->end() ) {
        Playlist * playlist = *iter;
        m_playlists->erase( iter );
        delete playlist;
    }

    // delete all tracks
    TrackList::ConstIterator trackIter = m_TrackList.const_iterator();
    while( trackIter.hasNext() ) {
        delete trackIter.next();
    }
    m_TrackList.clear();
    m_TrackIdMap.clear();

    itunesdbfile.setName(QString());
    timestamp = QDateTime();
    maxtrackid = 0;
    mMaxDBID = 0;
    // mainlist = itunesdb::Playlist();
    setDirty( false );
}


/*!
    \fn ITunesDB::removePlaylist( const QString& title)
 */
bool ITunesDB::removePlaylist( const QString& title, bool delete_instance) {
    ITunesDB::PlaylistContainer::iterator pos =
        std::find_if( m_playlists->begin(), m_playlists->end(), PlaylistByTitleFinder( title ) );

    if ( pos != m_playlists->end() ) {
        ITunesDBPlaylist * playlist = *pos;

        m_playlists->erase( pos );

        if ( delete_instance ) {
            delete playlist;
        }

    } else {
        return false;
    }

    setDirty();

    return true;
}


void ITunesDB::updateSmartPlaylists() {
    PlaylistContainer::iterator pos = m_playlists->begin();
    for ( ; pos != m_playlists->end(); ++pos ) {
        (*pos)->updateSmartPlaylist();
    }
}


/*!
    \fn ITunesDB::removeTrack(Q_UINT32 trackid, bool delete_instance = true)
 */
Q_UINT32 ITunesDB::removeTrack( Q_UINT32 trackid, bool delete_instance ) {
    TrackIDMap::const_iterator trackIdx = m_TrackIdMap.find( trackid );
    if ( trackIdx == m_TrackIdMap.end() ) {
        return 0;
    }

    uint idx = trackIdx->second;

    // remove track from track table
    m_TrackIdMap.erase( trackid );

    if ( idx >= m_TrackList.count() ) {
        return 0;
    }

    Track * track = m_TrackList[ idx ];
    m_TrackList[ idx ] = NULL;

    // remove track from playlists
    removeFromAllPlaylists( trackid );

    if( delete_instance ) {
        delete track;
    }

    setDirty();

    return trackid;
}


bool ITunesDB::renameAlbum(const QString& artist, const QString& title, const QString& newartist, const QString& newtitle) {
    // check some preconditions
    if ( newartist.isEmpty() && newtitle.isEmpty() ) {
        return false;   // doesn't make sense
    }
    if ( containsAlbum( newartist, newtitle ) ) {
        return false;
    }
    // update track info
    TrackPredicates::ByAlbum albumFinder(artist, title);
    ItDBTrackIter albumIter = m_TrackList.filteredElements(
                TrackPredicates::PredicatePtrDelegator( &albumFinder ) );
    if ( !albumIter.hasNext() ) {
        return false;   // no tracks with such artist/album found
    }
    while ( albumIter.hasNext() ) {
        itunesdb::Track * track = albumIter.next();
        if( track == NULL ) {
            continue;
        }
        if ( !newartist.isEmpty() ) {
            track->setArtist( newartist );
        }
        if ( !newtitle.isEmpty() ) {
            track->setAlbum( newtitle );
        }
    }

    setDirty();

    return true;
}


/**
 * Sets the title of the mainlist. 
 */
void ITunesDB::setTitle( const QString& title ) {
    m_Title = title;
    setDirty();
}


void ITunesDB::setDirty( bool dirtyFlag ) {
    if ( mDoDirtyFlagHandling ) {
        m_dirty = dirtyFlag;
        if ( dirtyFlag ) {
            mInternalVersion++;
        }
    }
}


bool ITunesDB::isDirty() const {
    return m_dirty;
}


/***************************************************************************
 *
 *        private/protected methods
 *
 ***************************************************************************/

void ITunesDB::removeFromAllPlaylists( Q_UINT32 trackid ) {
    PlaylistContainer::iterator pos = m_playlists->begin();
    for ( ; pos != m_playlists->end(); ++pos ) {
        (*pos)->removeAll( trackid );
    }
}

void ITunesDB::lock(bool write_lock) {
    if(!itunesdbfile.isOpen())
        itunesdbfile.open(IO_ReadOnly);

#ifdef WIN32
#warning TODO implement DB file locking
#else
    if(write_lock)
        flock(itunesdbfile.handle(), LOCK_EX);
    else
        flock(itunesdbfile.handle(), LOCK_SH);
#endif

    locked = true;
}

void ITunesDB::unlock() {
#ifdef WIN32
#warning TODO implement DB file locking
#else
    flock(itunesdbfile.handle(), LOCK_UN);
#endif
    itunesdbfile.close();
    locked = false;
}

bool ITunesDB::isLocked() const {
    return locked;
}

ITunesDBTrack * ITunesDB::insertTrackToDataBase( ITunesDBTrack * track ) {
    if ( getMHITTrackPosFor( track->getID() ) != -1 ) {
        // a track with this ID already exists
        return NULL;
    }

    if ( maxtrackid < track->getID() ) {
        maxtrackid = track->getID();
    }

    if ( mMaxDBID < track->getDBID() ) {
        mMaxDBID = track->getDBID();
    }

    // some checks
    if( track->getArtist().isEmpty() ) {
        track->setArtist( "_no_artist_" );
    }
    if( track->getAlbum().isEmpty() ) {
        track->setAlbum( "_no_album_" );
    }
    if( track->getTitle().isEmpty() ) {
        track->setTitle( "_no_title_" );
    }

    // set dirty flag to false when reading the itunesdb
    // will be set to true when using addTrack()
    track->setDirty( false );

    // synchronize this when making it thread safe - begin
    m_TrackList.append( track );
    m_TrackIdMap[ track->getID() ] = m_TrackList.count() - 1;

    return track;
}


Q_UINT32 ITunesDB::getNumTracks() const {
    return m_TrackIdMap.size();
}


Q_UINT32 ITunesDB::getNumPlaylists() const {
    return m_playlists->size();
}

Q_UINT32 ITunesDB::getNextFreeTrackID() const {
    Q_UINT32 trackid = maxtrackid;
    while ( getTrackByID( ++trackid ) != NULL ) { // look for the next free number
#if 0
    	printf( "getting next free trackid\n" );
#endif
        if ( trackid >= PLAYLISTITEM_INVALID ) trackid = 0x7E;    // start anew when max tracknumber is reached
    }
    return trackid;
}


Q_UINT64 ITunesDB::getNextFreeDBID() const {
    Q_UINT64 result = mMaxDBID + 2;
    while ( findTrackByDBID( result ) || getPlaylistByID( result ) ) {
        result += 2;
    }
    return mMaxDBID + 2;
}


#if QT_VERSION >= 0x040000
#define absolutePathQT43(dir) (dir.absolutePath())
#define splitQStringQT43(string,delim) string.split( delim, QString::SkipEmptyParts )
#else
#define absolutePathQT43(dir) (dir.absPath())
#define splitQStringQT43(string,delim) QStringList::split( string, delim, false )
#endif

QString ITunesDB::getFileForPathInfo( QString pathinfo ) const {
    QDir baseDir( getIPodBase() );
#if 0
    printf( "getFileForPathInfo( %s )\n", pathinfo.ascii() );
#endif
    QFile realPath( baseDir.path() + pathinfo.replace( ":", "/") );
    if ( !realPath.exists() ) {
#if 0
        fprintf( stderr, "file %s doesn't exist - trying to fix path.\n", realPath.name().ascii() );
#endif
        // umm, try to fix the path the simple way and let the caller
        // handle the error if this fails
    	QStringList elements = splitQStringQT43( pathinfo, "/" );
    	if ( elements.count() > 3 ) {
        QDir musicDir( m_ipodSysInfo.getMusicDir() );
	    	if ( !musicDir.cd( elements[ 2 ] ) && !musicDir.cd( elements[ 2 ].lower() ) ) {
	    		musicDir.cd( elements[ 2 ].upper() );
	    	}
#if 0
	    	printf( "file directory is %s\n", musicDir.absolutePath().ascii() );
#endif
	    	realPath.setName( absolutePathQT43( musicDir ) + "/" + elements[ 3 ] );
    	}
    }

    return realPath.name();
}


/***
 *
 * TrackConstIterator implementation
 *
 ***/

