 /*****************************************************************************
 *   Copyright (C) 2006 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include "itunesdb.h"
#include "itunesdbparser.h"
#include "itunesdbwriter.h"
#include "itunesdblistener.h"
#include "itunesdbdatasource.h"

#define ITUNESSD "iTunesSD"

using namespace itunesdb;


class ITunesDB::ItunesDBReaderAdapter : public itunesdb::ItunesDBListener {

public:
    ItunesDBReaderAdapter( ITunesDB& itunesdb ) :
        m_itunesdb( itunesdb ),
        m_currentDataSet( 0 ),
        m_recentlyPlayed ( itunesdb.createNewPlaylist( RECENTLY_PLAYED_LIST_NAME ) )
    {
        m_recentlyPlayed->setSortOrder( Playlist::SORTORDER_TIME_PLAYED );
    }

    virtual ~ItunesDBReaderAdapter() {}

    virtual Playlist * createNewPlaylist() {
        // printf( "creating new playlist\n" );
        return m_itunesdb.createNewPlaylist( QString::null );
    }

    virtual Track * createNewTrack() {
        return m_itunesdb.createEmptyTrack( 0 );
    }

    virtual void parseStarted() {}
    virtual void parseFinished() {}

    virtual void handleTrack( Track * track ) {
        ITunesDBTrack * dbtrack = dynamic_cast<ITunesDBTrack *>( track );
        if ( !dbtrack ) {
            delete track;
            return;
        }

        m_itunesdb.insertTrackToDataBase( dbtrack );
    }

    // TODO move playcounts and otg to different listeners/readers
    virtual void handlePlaycount(
            Q_UINT32 idx, Q_UINT32 lastplayed, Q_UINT32 stars, Q_UINT32 count, Q_UINT32, Q_UINT32 skipcount, Q_UINT32 last_skipped) {

        Track * track = m_itunesdb.getTrackAt( idx ); //Same trick as in the OTG Playlists this is an offset
        if ( !track ) {
            return;
        }

        if ( skipcount != 0 || last_skipped > track->getLastSkippedTime() ) {
            // printf( "%s got skipped at %u now %u\n", track->toQString().ascii(), last_skipped, QDateTime::currentDateTime().toTime_t() );
            track->setSkipCount( track->getSkipCount() + skipcount );
            track->setLastSkippedTime( last_skipped );
        }

        if ( count != 0 || lastplayed != track->getLastPlayed() ) {
            track->setLastPlayed( lastplayed );
            track->setPlayCount( track->getPlayCount() + count );
            track->setRecentlyPlayed();

            // Now insert track into the sorted List
            m_recentlyPlayed->addPlaylistItem( track->getID() );
        }

        // track got really changed
        if ( stars != 0 && track->getRating() != stars ) {
            track->setRating( (unsigned char) stars );
        }
    }

    void convertOffsetsToIDs( Playlist * playlist ) {
        Q_UINT32 idx;

        if ( !m_itunesdb.m_TrackList.count() ) {
            return;    // Something has gone VERY wrong here 
        }

        playlist->clear();

        for (uint i = 0; i < playlist->getNumTracks(); i++) {
            idx = playlist->getTrackIDAt( i ); // Get the ID the playlist uses
            if ( idx >= m_itunesdb.m_TrackList.count() ) {
                continue;
            }
            Track * track = m_itunesdb.m_TrackList[ idx ];
            if ( track != NULL ) {
                playlist->addPlaylistItem( track->getID() );
            }
        }
    }

    // TODO move playcounts and otg to different listeners/readers
    virtual void handleOTGPlaylist( Playlist * playlist ) {
        std::auto_ptr<Playlist> plCleanup( playlist );
        QString title;

        if ( !m_itunesdb.m_TrackList.count() ) {
            return;    // Something has gone VERY wrong here
        }

        if ( !playlist->getNumTracks() ) {
            return;
        }

        convertOffsetsToIDs( playlist );

        ITunesDBPlaylist * newplaylist = dynamic_cast<ITunesDBPlaylist *>( playlist );
        if ( !newplaylist ) {
            // printf( "Couldn't cast playlist to itunesdbplaylist\n" );
            return;
        }

        for ( uint i = 1; i < 100 && title.isNull() ; ++i) {
            title = QString("%1 %2").arg( "OTG Playlist" ).arg( QString::number( i ) );
            if ( m_itunesdb.getPlaylistByTitle( title ) ) {
                title = itunesdb::utils::NullQString;
            }
        }

        if (title.isNull()) {
            return;
        }

        newplaylist->setTitle( title );
        if ( m_itunesdb.addPlaylist( newplaylist ) ) {
            plCleanup.release();
        }
    }

    virtual void handlePlaylist( Playlist * playlist ) {
        std::auto_ptr<Playlist> plCleanup( playlist );

        // printf( "listener handling playlist %s\n", playlist->getTitle().ascii() );

        // This has to be before the podcast check since we really need the FIRST playlist after
        // the track section no matter if it's in the podcast block or not
        // TODO find out another way to find out if this is the mainlist (maybe a handleMainlist() or have some state thingy

        if ( playlist->isMasterPL() ) {
            if ( m_itunesdb.getTitle().isEmpty() ) {
                m_itunesdb.setTitle( playlist->getTitle() );
                // that's all we wanna know for now
            }

            return;
        }

        if ( m_currentDataSet != 2 || playlist->containsItemProperty( MHOD_PLAYLIST ) ) {
            m_itunesdb.hasPodcastsFlag |= playlist->getNumTracks() &&
                    playlist->isPodcastList();

            // can't handle podcasts at the moment
            return;
        }

        ITunesDBPlaylist * itunesdbplaylist = dynamic_cast<ITunesDBPlaylist *>( playlist );
        if ( itunesdbplaylist && m_itunesdb.addPlaylist( itunesdbplaylist ) ) {
            plCleanup.release();
        }
    }

    virtual void handleError(const QString &message) {
        m_itunesdb.error = message;
    }

    virtual void handleDataSet( Q_UINT32 type ) {
        m_currentDataSet = type;
    }

    virtual void setNumPlaylists(Q_UINT32) {}
    virtual void setNumTracks(Q_UINT32) {}

    ITunesDB& m_itunesdb;
    Q_UINT32 m_currentDataSet;

    // helper class to sort playcounts by last played date
    ITunesDBPlaylist * m_recentlyPlayed;
};


inline void updateTrackCount( TrackPtrList& album ) {
    TrackPtrList::Iterator albumIter = album.iterator();
    while( albumIter.hasNext() ) {
        albumIter.next()->setTrackCount( album.count() );
    }
    album.clear();
}


inline void updateTrackCountPerAlbum( const ITunesDB::TrackList& masterTrackList ) {
    TrackPtrList allTracks( masterTrackList.iterator() );
    allTracks.sort();

    QString lastAlbum;
    QString lastArtist;
    TrackPtrList album;

    TrackPtrList::ConstIterator trackIter = allTracks.const_iterator();
    while ( trackIter.hasNext() ) {
        Track * track = trackIter.next();

        if ( track->getAlbum().compare( lastAlbum ) != 0 || track->getArtist().compare( lastArtist ) != 0 ) {
            updateTrackCount( album );
            lastAlbum = track->getAlbum();
            lastArtist = track->getArtist();
        }

        album.append( track );
    }

    updateTrackCount( album );
}


bool ITunesDB::open() {
    error = itunesdb::utils::NullQString;

    mDoDirtyFlagHandling = false;
    mInternalVersion = 0;
    maxtrackid = 0; // be sure to get the lowest possible trackid given

    QString itunesDir( m_ipodSysInfo.getDeviceMountpoint() + "/iPod_Control/iTunes/" );

    itunesdbfile.setName( itunesDir + "iTunesDB" );
    itunessdfile.setName( itunesDir + ITUNESSD );

    ItunesDBReaderAdapter listener( *this );
    ItunesDBParser parser( listener );

    if( itunesdbfile.exists() ) {
        parser.parse( itunesdbfile );
        timestamp = QFileInfo( itunesdbfile ).lastModified();
    } else {
        return false;
    }

    // *** some post-parsing checks
    if ( m_Title.isEmpty()) {
        m_Title = "kpod";
    }

    if ( mMaxDBID == 0 ) {
        // generate starting dbid
        srand( getNumTracks() + getNumPlaylists() + 1 );
        do {
            mMaxDBID = ((Q_UINT64)rand()) << 32;
            mMaxDBID |= rand();
        } while( mMaxDBID > 0xAFFFFFFFFFFFFFFFLL ); // just to be really sure ...
    }


    // check for Track Metadata
    TrackList::ConstIterator trackIter = m_TrackList.const_iterator();
    while( trackIter.hasNext() ) {
        ITunesDBTrack * track = trackIter.next();
        if ( !track ) continue;
        if ( track->getDBID() == 0 ) {
            Q_UINT64 nextDBID = getNextFreeDBID();
            track->setDBID( nextDBID );
            mMaxDBID = nextDBID;
        }
    }

    if ( maxtrackid == 0 ) {
        // obviously no tracks where in the DB, set our starting point for new IDs
        maxtrackid = 0x7E;  // will be incremented for the first new track
    }

    // check playlists
    PlaylistContainer::iterator playlist = m_playlists->begin();
    for( ; playlist != m_playlists->end(); ++playlist ) {
        (*playlist)->doPostParsingChecks();
        if ( (*playlist)->getID() == 0 ) {
            Q_UINT64 nextDBID = getNextFreeDBID();
            (*playlist)->setID( nextDBID );
            mMaxDBID = nextDBID;
        }
    }

    // update numtracksinAlbum properties in all tracks
    updateTrackCountPerAlbum( m_TrackList );

    // already set to true here since we may need to update smart playlists
    mDoDirtyFlagHandling = true;

    QDir dir( itunesDir );
    dir.setNameFilter( "OTGPlaylistInfo*" );
    for ( unsigned int i = 0; i < dir.count(); i++ ) {
        if (QFileInfo(dir.filePath( dir[i] ) ).size()) {
            QFile file( dir.filePath( dir[i] ) );
            parser.parseOTG( file );
        }
    }

    QFile myfile ( itunesDir + "Play Counts" );
    if (myfile.exists()) {
        parser.parsePlaycount( myfile );
        ITunesDBPlaylist * recentlyPlayed = listener.m_recentlyPlayed;

        if ( recentlyPlayed->getNumTracks() ) {
            removePlaylist( RECENTLY_PLAYED_LIST_NAME , true);

            m_playlists->push_back( recentlyPlayed );
        } else {
            delete recentlyPlayed;
        }
    } else {
        delete listener.m_recentlyPlayed;
    }

    setDirty( false );

    return true;
}



class ITunesDB::ItunesDBWriterAdapter : public itunesdb::ItunesDBDataSource {
    const ITunesDB& m_itunesdb;
    bool m_writeSuccess;
    TrackPtrList m_trackList;

    ITunesDB::PlaylistContainer::iterator m_currentPL;

public:

    QString errorMsg;

    ItunesDBWriterAdapter( const ITunesDB& itunesdb )
        :   m_itunesdb( itunesdb ), m_writeSuccess( false ),
            m_trackList(),
            m_currentPL( m_itunesdb.m_playlists->begin() )
    {
        m_itunesdb.getAllTracks( m_trackList );
    }

    virtual ~ItunesDBWriterAdapter() {
    }

    /******************************************************
    *
    *       ItunesDBDataSource Methods
    *
    * for documentation of the following methods
    * see itunesdb/itunesdbdatasource.h
    *****************************************************/

    void writeInit() {
    }

    void writeFinished() {
        m_writeSuccess = true;
    }

    Q_UINT32 getNumPlaylists()
    {
        return m_itunesdb.m_playlists->size();
    }

    void handleError( const QString &message ) {
        errorMsg = message;
    }

    Q_UINT32 getNumTracks() {
        return m_itunesdb.m_TrackList.count();
    }

    TrackPtrList& getTrackList() {
        return m_trackList;
    }

    Playlist * firstPlaylist()
    {
        m_currentPL = m_itunesdb.m_playlists->begin();
        if ( m_currentPL != m_itunesdb.m_playlists->end() ) {
            return *m_currentPL;
        }
        return NULL;
    }

    Playlist * nextPlaylist()
    {
        if (   m_currentPL == m_itunesdb.m_playlists->end() ||
             ++m_currentPL == m_itunesdb.m_playlists->end() ) {
            return NULL;
        }
        return *m_currentPL;
    }

    const QString& getTitle() {
        return m_itunesdb.getTitle();
    }

    bool writeSuccessful() const {
        return m_writeSuccess;
    }
};

#if QT_VERSION >= 0x040000
#define absoluteFilePathQT43(dir, file) (dir.absoluteFilePath(file))
#else
#define absoluteFilePathQT43(dir, file) (dir.absFilePath(file))
#endif

// check if the new written ITunesDB file is consistent with the given database
inline bool isDBConsistent( const ITunesDB& origDB, ITunesDB& newDB, QString& perror ) {
	if ( !newDB.open() ) {
	    perror = "DB Consistency Check: could not read DB.";
		return false;
	}
	
	if ( newDB.getTitle() != origDB.getTitle() ) {
        perror = "DB Consistency Check: Title doesn't match.";
		return false;
	}
	
	if ( newDB.getNumTracks() != origDB.getNumTracks() ) {
        perror = "DB Consistency Check: Number of tracks differs.";
		return false;
	}
	
	ITunesDB::TrackConstIterator origTrackIter = origDB.getAllTracks();
	ITunesDB::TrackConstIterator newTrackIter = newDB.getAllTracks();
	while ( newTrackIter.hasNext() ) {
		if ( !origTrackIter.hasNext() ) {
			return false;
		}

		Track * newTrack = newTrackIter.next();
		Track * origTrack = origTrackIter.next();
		if ( newTrack->getID() != origTrack->getID() ) {
	        perror = "DB Consistency Check: Track order differs.";
			return false;
		}
		if ( newTrack->getNumComponents() != origTrack->getNumComponents() ) {
            perror = QString( "DB Consistency Check: Track metadata differs for track %1 - %2 - %3." )
                        .arg( origTrack->getArtist() ).arg( origTrack->getAlbum() ).arg( origTrack->getTitle() );
			return false;
		}
	}
	
	if ( newDB.getNumPlaylists() != origDB.getNumPlaylists() ) {
        perror = "DB Consistency Check: Number of playlists differs.";
		return false;
	}
	
	return true;
}

// TODO hav a read only flag on the iTunesDB so we cannot change it
bool compatibleWith( QFile& oldItunesDBFile, QString& perror ) {
    Q_UINT32 needsHash;
    bool wasOpen = oldItunesDBFile.isOpen();
    if ( !wasOpen && !oldItunesDBFile.open( IO_ReadOnly ) ) {
    	perror = "Couldn't check itunesdbfile.";
        return true;    // not possible to check - ignore and overwrite
    }
    oldItunesDBFile.at( 48 );
    needsHash = oldItunesDBFile.getch();
    if ( needsHash ) {
        perror = "This version of libqtpod doesn't support writing the itunesdb hash value.";
    }
    
    if ( !wasOpen ) {
        oldItunesDBFile.close();
    }
    
    return !needsHash;
}

bool ITunesDB::writeDatabase(const QString& filename) {
    QFile outfile( filename );
    if( filename.isEmpty() ) {
        outfile.setName( itunesdbfile.name() );
    }

    // cleanups
    error = itunesdb::utils::NullQString;  // reset error msg

    // remove undefined/somehow_messed_up tracklist items
    removeFromAllPlaylists( PLAYLISTITEM_INVALID );

    updateSmartPlaylists();
    std::sort( m_playlists->begin(), m_playlists->end(),
        ListItem::LessThanStringPropertyComparator( MHOD_TITLE ) );

    QFileInfo outfileInfo( outfile );
    QDir dir( outfileInfo.dir( TRUE ) );

    // create a backup of the original database
    QString itunesdb_backupname = outfileInfo.fileName() + ".bak";
    if ( outfile.exists() ) {
        if ( !compatibleWith( outfile, error ) ) {
            return false;
        }
        // remove old backup copy - on some FS this is necessary
        dir.remove( itunesdb_backupname );
        dir.rename( outfileInfo.fileName(), itunesdb_backupname );
    }

    // write the new database
    ItunesDBWriterAdapter source( *this );
    ItunesDBWriter writer( source );
    writer.write(outfile);
    ITunesDB newDB( m_ipodSysInfo );
    if ( !source.writeSuccessful() || ( filename.isEmpty() && !isDBConsistent( *this, newDB, error ) ) ) {
        // restore backup
        dir.remove( outfileInfo.fileName() );
        dir.rename( itunesdb_backupname, outfileInfo.fileName() );
        return false;
    }

    // remove OnTheGo Playlist files
    dir.setNameFilter( "OTGPlaylistInfo*" );
    for ( unsigned int i = 0; i < dir.count(); i++ ) {
        if (QFileInfo(dir.filePath( dir[i] ) ).size()) {
            dir.remove( dir[i] );
        }
    }

    // create backup copy of the Play Counts file
    dir.remove( "Play Counts.bak" );
    dir.rename( "Play Counts", "Play Counts.bak" );

    // create a backup copy and write the ITunesSD file for ipod shuffle
    QString itunesSDname( QFileInfo( itunessdfile ).fileName() );
    if ( itunesSDname.isEmpty() ) {
        itunesSDname = absoluteFilePathQT43( dir, ITUNESSD );
    }
    if ( dir.exists( itunesSDname ) ) {
        QString backupname( itunesSDname + ".bak" );

        // remove old backup copy - on some FS this is necessary
        dir.remove( backupname );

        dir.rename( itunesSDname, backupname );
    }
    QFile outfilesd( dir.filePath( itunesSDname ) );
    writer.writeSD( outfilesd );

    error = source.errorMsg;
    if ( filename.isEmpty() ) {
        setDirty( false );    // container is in sync with the database now
    }

    return true;
}

