 /*****************************************************************************
 *   Copyright (C) 2006 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include "itunesdbparser.h"
#include "utils.h"

#include <qdatastream.h>
#include <qfileinfo.h>


namespace itunesdb {

inline Q_UINT32 toPCDate( Q_UINT32 macdate ) {
    if ( macdate ) {
        return macdate - MAC_EPOCH_DELTA;
    }
    return macdate;
}

class FileCloser {
    QFile & file;
public:
    FileCloser( QFile& toClose ) : file( toClose ) {}
    ~FileCloser() {
        file.close();
    }
};

ItunesDBParser::ItunesDBParser( ItunesDBListener& ituneslistener ) : current_playlist( NULL )
{
    listener = &ituneslistener;
}


ItunesDBParser::~ItunesDBParser()
{
}


/*!
    \fn itunesdb::ItunesDBParser::parse(QFile& file)
 */
void ItunesDBParser::parse( QFile& file ) {
    Q_UINT32 magic, headerlen, filesize;

    PlaylistItem * current_playlistitem = NULL;
    Track * current_track = NULL;

    ListItem * listitem = NULL;

    listener->parseStarted();

    if( !file.exists() || !file.open( IO_ReadOnly ) ) {
        listener->handleError( file.name() + " could not be opened!" );
        return;    // no need to cleanup
    }

    FileCloser closer( file );

    QDataStream stream( &file );
    stream.setByteOrder( QDataStream::LittleEndian );

    stream >> magic;
    if ( magic != 0x6462686D ) {
        listener->handleError( file.name() + " is not an itunesDB file" );
        return;
    }

    stream >> headerlen;
    stream >> filesize;

    // TODO dbID setzen

    // skip the rest of the header
    itunesdb::utils::seekRelative( stream, headerlen - 12 );

    // begin with actual parsing
    while( !stream.atEnd() ) {
        Q_UINT32 blocktype;
        Q_UINT32 blocklen;

        stream >> blocktype;

        switch( blocktype ) {
        case 0x6473686D: {    // mhsd - yeah, good to know
            Q_UINT32 type;
            stream >> blocklen;

            // iTunesDB has two large blocks - one with the tracks and one with playlists
            // both begin with this tag. Since the data entrys themself have different block
            // types we can safely ignore this information here ...
            if ( listitem != NULL ) {    // OK, so now we're at the end of the first mshd
                handleItem( *listitem );
                listitem = NULL;
            }
            if ( current_playlist ) {   // OK, we're at the end of a block containing playlists
                // there's a playlist waiting to be handed over to the listener
                current_playlist->setDirty( false );
                current_playlist->sort();
                listener->handlePlaylist( current_playlist );
                current_playlist = NULL;
            }
            stream >> type;     // skip mhsd size
            stream >> type;
            listener->handleDataSet( type );

            itunesdb::utils::seekRelative( stream, blocklen- 16 );
            }
            break;

        case 0x746C686D: {    // mhlt: number of tracks (and other metainfo?)
            Q_UINT32 numtracks;
            stream >> blocklen;
            if( listitem != NULL ) {
                handleItem( *listitem );
                listitem = NULL;
            }
            stream >> numtracks;
            listener->setNumTracks( numtracks );
            itunesdb::utils::seekRelative( stream, blocklen - 12 );    // skip the rest
            }
            break;

        case 0x7469686D: {    // mhit - start of a track
            if ( listitem != NULL ) {
                handleItem( *listitem );
            }

            current_track = listener->createNewTrack();
            bool ok = false;
            current_track->readFromStream( stream, &ok );
            if ( !ok ) {
                delete current_track;
                current_track = NULL;
            }
            listitem = current_track;
            }
            break;

        case 0x646F686D:      // mhod
            Q_UINT32 blocklen;
            Q_UINT32 type;
            stream >> blocklen;
            stream >> blocklen;
            stream >> type;
            if ( listitem != NULL ) {
                listitem->readMHOD( stream, type, blocklen );
            } else {
                itunesdb::utils::seekRelative( stream, blocklen - 16 );
            }
            break;
        case 0x706C686D: {    // mhlp
            Q_UINT32 numplaylists;
            if ( listitem != NULL ) {
                handleItem( *listitem );
                listitem = NULL;
            }
            stream >> blocklen;
            stream >> numplaylists;
            listener->setNumPlaylists( numplaylists );
            itunesdb::utils::seekRelative( stream, blocklen - 12 );
            }
            break;

        case 0x7079686D: {    // mhyp: Playlist
            if( listitem != NULL ) {
                handleItem( *listitem );
            }

            if( current_playlist ) {
                // there's a playlist waiting to be handed over to the listener
                current_playlist->setDirty( false );
                current_playlist->sort();
                listener->handlePlaylist( current_playlist );
            }
            current_playlist = listener->createNewPlaylist();

            bool ok = false;
            current_playlist->readFromStream( stream, &ok );
            if ( !ok ) {
                delete current_playlist;
                current_playlist = NULL;
            }

            listitem = current_playlist;
            }
            break;
        case 0x7069686D: {    // mhip: PlaylistItem
            Q_UINT32 itemid;

            stream >> blocklen;
            if( listitem != NULL ) {
                handleItem( *listitem );
            }

            itunesdb::utils::seekRelative( stream, 16 );
            stream >> itemid;

            current_playlistitem = NULL;

            if ( current_playlist ) {
                current_playlistitem = current_playlist->createNewItem( itemid );
                listitem = current_playlistitem;
            }

            itunesdb::utils::seekRelative( stream, blocklen - 28 );
            }
            break;
        case 0x616C686D:    // mhla
            // ignore this record
            // TODO add a callback function for this in the listener interface
            stream >> blocklen;
            itunesdb::utils::seekRelative( stream, blocklen - 8 );
            break;
        case 0x6169686D:    // mhia
            // ignore this and all its childs
            // TODO add it to the list of mhia records controlled by the corresponding mhla
            stream >> blocklen;
            stream >> blocklen;
            itunesdb::utils::seekRelative( stream, blocklen - 12 );
            break;
        default: {
            QString msg( "unknown tag %1 found at position %2 ! stop parsing" );
            listener->handleError( msg.arg( blocktype, 0, 16 ).arg( file.at() ) );

            delete current_playlist;
            current_playlist = NULL;

            delete current_playlistitem;
            current_playlistitem = NULL;

            delete current_track;
            current_track = NULL;

            }

            return;
        }
    }

    if ( listitem != NULL ) {
        handleItem( *listitem );
        listitem = NULL;
    }
    if ( current_playlist ) {
        // there's a playlist waiting to be handed over to the listener
        current_playlist->setDirty( false );
        listener->handlePlaylist( current_playlist );
        current_playlist = NULL;
    }
    current_playlistitem = NULL;
    listener->parseFinished();
}

void ItunesDBParser::parsePlaycount( QFile& file ) {
    Q_UINT32 magic, numtracks, entrysize, headersize;
    Q_UINT32 data[ 7 ] = { 0, 0, 0, 0, 0, 0, 0 }; //Items in the Entrys

    if ( !file.exists() || !file.open( IO_ReadOnly ) ) {
        // listener->handleError( file.name()+ " could not be opened!"); // this should be optional
        return;    // no need to cleanup
    }

    FileCloser closer( file );

    QDataStream stream( &file );
    stream.setByteOrder( QDataStream::LittleEndian );

    stream >> magic;
    if( magic != ITUNESDB_MHDP ) { //Needed this defines, otherwise my eyes hurt ;)
        // listener->handleError( file.name() + " is not an Playcount File");	// this should be optional
        return;
    }

    stream >> headersize;
    stream >> entrysize;
    stream >> numtracks;
    if ( entrysize < 0x0C ) {
        // listener->handleError( "You use a old firmware version, go get a new one :-)"); // this should be optional
        return;
    }
    if ( QFileInfo( file ).size() != ( headersize + entrysize * numtracks ) ) {
        // listener->handleError( file.name() + " is corrupt"); //Actually this is not an error, but a bad sign for the hd
        //listener->handleError( "File System error, kiss your data goodbye as long as you could!"); 
        return;
    }

    itunesdb::utils::seekRelative( stream, headersize - 16 );    // skip empty header
    //This i is (@see OTG Playlists) an INDEX in the mhdxtxft ;)
    for ( uint i = 0; i < numtracks; i++ ) {
        for ( uint j = 0; ( j * 4 ) < entrysize && j < 7 ; j++ ) {
            stream >> data[ j ];
        }
        // check dates
        data[ 1 ] = toPCDate( data[ 1 ] );
        data[ 6 ] = toPCDate( data[ 6 ] );

        listener->handlePlaycount( i, data[ 1 ], data[ 3 ], data[ 0 ], data[ 2 ], data[ 5 ], data[ 6 ] );

        if ( entrysize > 28 ) {
            itunesdb::utils::seekRelative( stream, entrysize - 28 );    // skip unknown 
        }
    }
}

void ItunesDBParser::parseOTG( QFile& file ) {
    Q_UINT32 magic, numtracks, filesize;

    if( !file.exists() || !file.open( IO_ReadOnly ) ) {
        listener->handleError( file.name() + " could not be opened!" );
        return;    // no need to cleanup
    }

    FileCloser closer( file );

    QDataStream stream( &file );
    stream.setByteOrder( QDataStream::LittleEndian );

    stream >> magic;
    if ( magic != 0x6F70686D ) { //"mhpo" files start with this
        listener->handleError( file.name() + " is not an OTG file" );
        return;
    }

    current_playlist = listener->createNewPlaylist();   // Make a shiny new playlist, note: Does not get a name
    current_playlist->setTitle( "kpod ruleZ!" ); //will be overwritten by handleOTGPlaylist

    stream >> filesize;
    stream >> magic; //Unknown entry, but interesing to know
    stream >> numtracks;
    stream >> magic; //Unknown entry, but interesing to know

    //Hmmm the index returned here are not normal indizes but offsets in the MODT (?) 
    int idx;
    for ( unsigned int i = 0; i < numtracks; i++ ) {
        stream >> idx;
        PlaylistItem * item = current_playlist->createNewItem( idx );
        handleItem( *item );
    }
    listener->handleOTGPlaylist( current_playlist );

    current_playlist = NULL;
    //listener->parseFinished();
}

}

/*!
    \fn itunesdb::ItunesDBParser::handleItem( ListItem& item)
 */
void itunesdb::ItunesDBParser::handleItem( ListItem& item )
{
    switch ( item.getType() ) {
    case ITEMTYPE_TRACK: {
        Track * pTrack = dynamic_cast <Track *> ( &item );
        if( pTrack == NULL || pTrack->getID() == 0 ) {
            break;
        }
        pTrack->doneAddingData();
        listener->handleTrack( pTrack );
        }
        break;
    case ITEMTYPE_PLAYLIST: {
        Playlist * pPlaylist = dynamic_cast <Playlist *> ( &item );
        if ( pPlaylist == NULL ) {
            break;
        }
        pPlaylist->doneAddingData();
        }
        break;
    case ITEMTYPE_PLAYLISTITEM: {
        PlaylistItem * pPlaylistitem = dynamic_cast <PlaylistItem *> ( &item );
        if( pPlaylistitem == NULL ) {
            break;
        }
        pPlaylistitem->doneAddingData();
        if ( current_playlist ) {
            current_playlist->addPlaylistItem( pPlaylistitem );
        }
        }
        break;
    }
}

