 /*****************************************************************************
 *   Copyright (C) 2004 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include "itunesdbwriter.h"

#include <qdatastream.h>
#include <qbuffer.h>
#include <qtextcodec.h>

#include <algorithm>

#ifdef WIN32
#include <Windows.h>
#else
#include <unistd.h> // for fsync()
#endif

#if QT_VERSION >= 0x040000
#define ptrWhenQT4(a) (&a)
#else
#define ptrWhenQT4(a) a
#endif

#define MHSD_HEADERLEN 0x60

namespace itunesdb {

ItunesDBWriter::ItunesDBWriter(ItunesDBDataSource & trackdatasource) : m_datasource( trackdatasource ) {}

ItunesDBWriter::~ItunesDBWriter() {}


/*!
    \fn itunesdb::ItunesDBWriter::write(QFile& file)
 */
void ItunesDBWriter::write(QFile& file)
{
    QByteArray trackdata;
    QByteArray playlistdata;

    if ( !file.open( IO_WriteOnly) ) {
        m_datasource.handleError( file.name() + " could not be opened for writing!" );
        return;
    }

    m_datasource.writeInit();

    fillTrackBuffer( trackdata);
    fillPlaylistBuffer( playlistdata);

    // write to file
    QDataStream stream( &file);
    stream.setByteOrder( QDataStream::LittleEndian);

    stream << (Q_UINT32) 0x6462686D;    // write magic = mhbd
    stream << (Q_UINT32) 0x68;   // headerlen = 104 byte
    stream << (Q_UINT32) 0x0;    // length of the whole file - set at the end
    stream << (Q_UINT32) 0x1;
    stream << (Q_UINT32) 0x13;   // iTunes version 7.0
    stream << (Q_UINT32) 0x2;    // we don't write podcasts at the moment (4.9 can have a third for podcasts )
    stream << (Q_UINT64) 0x0;    // unkown 64 bit 
    stream << (Q_UINT32) 0x2;
    stream << (Q_UINT64) 0x0;

    QString localename( QTextCodec::locale() );
    if ( localename.length() < 2 ) {
        localename = "en";
    }
    stream.writeRawBytes( localename, 2 );

    for( int i= 0; i< 29; i++) {
        stream << (Q_UINT16) 0;
    }

    // write track data
    stream.writeRawBytes( trackdata.data(), trackdata.size());
    
    // write playlist data
    stream.writeRawBytes( playlistdata.data(), playlistdata.size());
    
    file.at( 8);
    stream << (Q_UINT32) file.size();
    
    stream.unsetDevice();
    file.flush();

    #ifdef WIN32
    FlushFileBuffers( (HANDLE)file.handle() );
#else
    fsync( file.handle() );
#endif
    
    file.close();

    m_datasource.writeFinished();
}


/*!
    \fn itunesdb::ItunesDBWriter::writeSD(QFile& file)
 */
void itunesdb::ItunesDBWriter::writeSD( QFile& file )
{
    if( !file.open( IO_WriteOnly)) {
        m_datasource.handleError(file.name()+ " could not be opened for writing!");
        return;
    }

    // write to file
    QDataStream stream( &file );
    stream.setByteOrder( QDataStream::BigEndian );

    //stream << (Q_UINT8) 0 << (Q_UINT16) m_datasource.getNumTracks(); // limited to 64k songs
    //stream << (Q_UINT8) 0x01 << (Q_UINT16) 0x0600;
    //stream << (Q_UINT8) 0x00 << (Q_UINT16) 0x0012;
    write3ByteLittle( stream, m_datasource.getNumTracks() );
    write3ByteLittle( stream, 0x010600 );
    write3ByteLittle( stream, 0x000012 );
    for( int i= 0; i< 9; i++)
        stream << (Q_UINT8) 0;

    QByteArray trackdata;
    fillTrackBufferSD( trackdata );
    // write track data
    stream.writeRawBytes( trackdata.data(), trackdata.size());

    file.close();
}

void itunesdb::ItunesDBWriter::fillTrackBuffer( QByteArray& buffer)
{
    QBuffer io_buffer( ptrWhenQT4(buffer) );
    io_buffer.open(IO_WriteOnly);
    QDataStream stream( &io_buffer);
    stream.setByteOrder( QDataStream::LittleEndian);

    // write mhsd
    stream << (Q_UINT32) 0x6473686D;    // "mhsd"
    stream << (Q_UINT32) MHSD_HEADERLEN;    // headerlen
    stream << (Q_UINT32) 0x0;    // length - set when we're done
    stream << (Q_UINT32) 1;        // type: tracklist
    for( int i= 0; i< 20; i++)
        stream << (Q_UINT32) 0;    // pad the rest (80 bytes)

    // tracklist header
    stream << (Q_UINT32) 0x746C686D;    // mhlt
    stream << (Q_UINT32) 0x5C;    // headerlen
    stream << (Q_UINT32) m_datasource.getNumTracks();    // number of tracks
    for( int i= 0; i< 20; i++)
        stream << (Q_UINT32) 0;    // pad the rest (80 bytes)

    // list the tracks
    TrackPtrList& trackPtrList( m_datasource.getTrackList() );
    TrackPtrList::Iterator trackIter = trackPtrList.iterator();
    while( trackIter.hasNext() ) {
        trackIter.next()->writeToStream( stream );
    }

    io_buffer.at( 8);
    stream << (Q_UINT32)io_buffer.size();    // write the length

    io_buffer.close();
}


/**
 *
void itunesdb::ItunesDBWriter::fillPodcastBuffer( QByteArray& buffer )
{
    QBuffer io_buffer( buffer);
    io_buffer.open(IO_WriteOnly);
    QDataStream stream( &io_buffer);
    stream.setByteOrder( QDataStream::LittleEndian);

    stream << (Q_UINT32) 0x6473686D;    // "mhsd"
    stream << (Q_UINT32) MHSD_HEADERLEN;    // headerlen
    stream << (Q_UINT32) 0x0;    // length - set when we're done
    stream << (Q_UINT32) 3;        // type: podcastlist
    for( int i= 0; i< 20; i++)
        stream << (Q_UINT32) 0;    // pad the rest (80 bytes)
    
    stream << (Q_UINT32) 0x706C686D;     // "mhlp"
    stream << (Q_UINT32) 0x5C;    // headerlen
    stream << (Q_UINT32) 1;          // one podlist
    for( int i= 0; i< 20; i++)
        stream << (Q_UINT32) 0;    // pad the rest (80 bytes)
    // ???
    
    io_buffer.at( 8);
    stream << (Q_UINT32)io_buffer.size();    // write the length
    
    io_buffer.close();
}*/


class MasterPlaylist : public itunesdb::Playlist {
    TrackPtrList & m_trackPtrList;
    uint m_trackCount;
    uint * m_indices;
public:
    MasterPlaylist( itunesdb::TrackPtrList & allTracks )
        : itunesdb::Playlist(),
          m_trackPtrList( allTracks ),
          m_trackCount( allTracks.count() ),
          m_indices( new uint [ m_trackCount ] )
    {
        TrackPtrList::Iterator trackIter = m_trackPtrList.iterator();
        while( trackIter.hasNext() ) {
            addPlaylistItem( trackIter.next()->getID() );
        }
        for ( uint i = 0; i < m_trackCount; ++i ) {
            m_indices[ i ] = i;
        }
    }
    virtual ~MasterPlaylist() {
        delete [] m_indices;
    }

    struct IndexComparator {
        TrackComparatorPtr m_comparator;
        TrackPtrList& m_trackList;
        IndexComparator( const TrackComparatorPtr& comparator, TrackPtrList& trackList )
            : m_comparator( comparator ), m_trackList( trackList ) {}
        virtual ~IndexComparator() {}

        bool operator()( const uint& i1, const uint& i2 ) {
            Track* track1 = m_trackList[ i1 ];
            Track* track2 = m_trackList[ i2 ];
            return m_comparator->compare( *track1, *track2 ) < 0;
        }
    };

    void sortIndices( const TrackComparatorPtr& comparator ) const {
        std::sort( &m_indices[ 0 ], &m_indices[ m_trackCount ], IndexComparator( comparator, m_trackPtrList ) );
    }

    void writeMHOD52( QDataStream& stream, uint sortIndex, const TrackComparatorPtr& comparator ) const {
        stream << (Q_UINT32) 0x646F686D;    // mhod
        stream << (Q_UINT32) 0x18;    // headerlen
        stream << (Q_UINT32) 72 + ( 4 * m_trackCount ); // complete size
        stream << (Q_UINT32) 52;
        stream << (Q_UINT32) 0;
        stream << (Q_UINT32) 0;
        stream << (Q_UINT32) sortIndex;
        stream << (Q_UINT32) m_trackCount;

        for ( uint i = 0; i < 10; i++ ) {
            stream << (Q_UINT32) 0;
        }

        sortIndices( comparator );

        for ( uint i = 0; i < m_trackCount; i++ ) {
            stream << (Q_UINT32) m_indices[ i ];
        }
    }

    virtual uint writeNonStringMHODs( QDataStream& outstream ) const {

        // write type52 MHODs
        writeMHOD52( outstream, 0x03, TrackComparators::BY_TITLE );
        writeMHOD52( outstream, 0x04, TrackComparators::BY_ALBUM );
        writeMHOD52( outstream, 0x05, TrackComparators::BY_ARTIST );
        writeMHOD52( outstream, 0x07, TrackComparators::BY_GENRE );
        writeMHOD52( outstream, 0x12, TrackComparators::BY_COMPOSER );

        // writeMHOD52( outstream, 0x1D, TrackComparators::BY_TVSHOW );
        // writeMHOD52( outstream, 0x1E, TrackComparators::BY_SEASON );
        // writeMHOD52( outstream, 0x1F, TrackComparators::BY_EPISODE );

        return 5;
    }
};

/*!
    \fn itunesdb::ItunesDBWriter::fillPlaylistBuffer( QByteArray& buffer)
 */
void itunesdb::ItunesDBWriter::fillPlaylistBuffer( QByteArray& buffer)
{
    QBuffer io_buffer( ptrWhenQT4(buffer) );
    io_buffer.open(IO_WriteOnly);
    QDataStream stream( &io_buffer);
    stream.setByteOrder( QDataStream::LittleEndian);

    MasterPlaylist masterlist( m_datasource.getTrackList() );
    masterlist.setTitle( m_datasource.getTitle() );

    // write mhsd
    stream << (Q_UINT32) 0x6473686D;    // "mhsd"
    stream << (Q_UINT32) MHSD_HEADERLEN;    // headerlen
    stream << (Q_UINT32) 0x0;    // length - set when we're done
    stream << (Q_UINT32) 2;        // type: playlist
    for( int i= 0; i< 20; i++)
        stream << (Q_UINT32) 0;    // pad the rest (80 bytes)

    // write mhlp
    stream << (Q_UINT32) 0x706C686D;    // "mhlp"
    stream << (Q_UINT32) 0x5C;    // headerlen
    stream << (Q_UINT32) m_datasource.getNumPlaylists()+ 1;    // number of playlists incl. mainlist
    for( int i= 0; i< 20; i++)
        stream << (Q_UINT32) 0;    // pad the rest (80 bytes)

    // write the mainlist
    masterlist.writeToStream( stream, true );

    for( Playlist * playlist= m_datasource.firstPlaylist(); playlist != NULL; playlist= m_datasource.nextPlaylist()) {
        playlist->writeToStream(stream, false);
    }

    io_buffer.at( 8);
    stream << (Q_UINT32)io_buffer.size();    // write the length

    io_buffer.close();
}


/*!
    \fn itunesdb::ItunesDBWriter::writeTracks( QDataStream& stream)
 */
void itunesdb::ItunesDBWriter::fillTrackBufferSD( QByteArray& buffer)
{
    QBuffer io_buffer( ptrWhenQT4(buffer) );
    io_buffer.open(IO_WriteOnly);
    QDataStream stream( &io_buffer);
    stream.setByteOrder( QDataStream::BigEndian);

    // list the tracks
    TrackPtrList& trackPtrList( m_datasource.getTrackList() );
    TrackPtrList::Iterator trackIter = trackPtrList.iterator();
    while( trackIter.hasNext() ) {
        Track * track = trackIter.next();

        write3ByteLittle( stream, 0x22e );
        write3ByteLittle( stream, 0x5aa501 );
        write3ByteLittle( stream, 0 ); // starttime
        write3ByteLittle( stream, 0 );
        write3ByteLittle( stream, 0 );
        write3ByteLittle( stream, 0 ); // stoptime
        write3ByteLittle( stream, 0 );
        write3ByteLittle( stream, 0 );
        write3ByteLittle( stream, 0x64 ); // volume=0
        write3ByteLittle( stream, 0x01 ); // TODO only MP3 currently
        write3ByteLittle( stream, 0x200 );

        // filename
        QString filename = track->getPathInfo();
        // make "/" out of ":"
        int idx = filename.find( ':' );
        while ( idx>-1 )
        {
            filename[idx] = '/';
            idx = filename.find( ':' );
        }

        const char *data= (const char *)filename.ucs2();
        int datalen = 0;
        if( data != NULL)
        {
            datalen = 2* filename.length();
            stream.writeRawBytes( data, datalen );
        }
        int padlength = 522 - datalen;
        for ( int i=0; i<padlength; i++ )
            stream << (Q_UINT8) 0;

        stream << (Q_UINT8) 0x01;
        stream << (Q_UINT8) 0;
        stream << (Q_UINT8) 0;
    }
    io_buffer.close();
}


/**
 */
void ItunesDBWriter::write3ByteLittle( QDataStream &stream, int v )
{
    stream << (Q_UINT8) ((v>>16) & 0xff);
    stream << (Q_UINT8) ((v>>8) & 0xff);
    stream << (Q_UINT8) (v & 0xff);
}


}
