 /*****************************************************************************
 *   Copyright (C) 2004 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include "listitem.h"

#include "utils.h"

namespace itunesdb {

/**
 * @brief Returns true if the given Property represents a String Property
 * @param property the property to check
 * @return true if the given Property represents a String Property
 */
bool isStringPorperty( ItemProperty property ) {
	switch ( property ) {
    case MHOD_TITLE:
    case MHOD_PATH:
    case MHOD_ALBUM:
    case MHOD_ARTIST:
    case MHOD_GENRE:
    case MHOD_FDESC:
    case MHOD_EQ_SETTING:
    case MHOD_COMMENT:
    case MHOD_CATEGORY:
    case MHOD_COMPOSER:
    case MHOD_GROUPING:
    case MHOD_DESCRIPTION_TEXT:
    case MHOD_SUBTITLE:
    // TODO check for these with the libqpod impl
    // case MHOD_TVSHOW:
    // case MHOD_TVEPISODE:
    // case MHOD_TVNETWORK:
    case MHOD_ALBUMARTIST:
    case MHOD_ARTIST_FOR_SORT:
    case MHOD_KEYWORDS:
    case MHOD_TITLE_FOR_SORT:
    case MHOD_ALBUM_FOR_SORT:
    case MHOD_ALBUMARTIST_FOR_SORT:
    case MHOD_COMPOSER_FOR_SORT:
    	return true;
    default:
    	return false;
	}
}

struct DumpedMHODFinder {
    Q_UINT32 m_type2find;
    DumpedMHODFinder( Q_UINT32 type2find ) : m_type2find( type2find ) {}
    bool operator() ( const MHODPair& pair ) const {
        return pair.first == m_type2find;
    }
};

ListItem::ListItem(int type) : m_dirty( true ) {
    m_itemType = type;
}

ListItem::ListItem() : m_dirty( true ) {
    m_itemType = ITEMTYPE_NONE;
}

ListItem::~ListItem()
{
}

int ListItem::getType() const {
    return m_itemType;
}

void ListItem::setItemProperty(const QString& data, ItemProperty field) {
    if ( data.isEmpty() ) {
        m_properties.erase( field );
    } else {
        m_properties[ field ] = data;
    }
    setDirty();
}

const QString& ListItem::getItemProperty( ItemProperty field ) const {
    PropertyMap::const_iterator element = m_properties.find( field );
    if ( element != m_properties.end() ) {
        return (*element).second;
    }

    return itunesdb::utils::NullQString;
}


bool ListItem::containsItemProperty( ItemProperty field ) const {
    return m_properties.find( field ) != m_properties.end();
}


const QByteArray * ListItem::getDumpedMHOD( Q_UINT32 field ) const {

    DumpedMHODs::const_iterator dumpedElem = utils::findFirst(
        m_dumpedMHODs.begin(), m_dumpedMHODs.end(), DumpedMHODFinder( field ) );

    if ( dumpedElem != m_dumpedMHODs.end() ) {
        return &(*dumpedElem).second;
    }
    return NULL;
}


uint ListItem::getNumComponents() const {
    uint result = getNumStringComponents();
    for( DumpedMHODs::const_iterator dumpedElement = m_dumpedMHODs.begin();
         dumpedElement != m_dumpedMHODs.end(); ++dumpedElement) {
        if ( (*dumpedElement).second.size() > 0 ) {
            result++;
        }
    }
    return result;
}

uint ListItem::getNumStringComponents() const {
    uint result = 0;
    for( PropertyMap::const_iterator element = m_properties.begin(); element != m_properties.end(); ++element) {
        if ( !(*element).second.isEmpty() ) {
            result++;
        }
    }
    return result;
}

void ListItem::setDirty( bool dirtyFlag ) {
    m_dirty = dirtyFlag;
}

bool ListItem::isDirty() const {
    return m_dirty;
}


void ListItem::readNonStringMHOD( QDataStream& stream, Q_UINT32 type, Q_UINT32 blocklen ) {
    // dump the whole mhod
    uint numBytes = blocklen - 16;
    QByteArray buffer( numBytes );
    stream.readRawBytes( buffer.data(), numBytes );
    m_dumpedMHODs.push_back( MHODPair( type, buffer ) );
}


void ListItem::readMHOD( QDataStream& stream, Q_UINT32 type, Q_UINT32 blocklen ) {
    switch ( type ) {
    case MHOD_TITLE:
    case MHOD_PATH:
    case MHOD_ALBUM:
    case MHOD_ARTIST:
    case MHOD_GENRE:
    case MHOD_FDESC:
    case MHOD_EQ_SETTING:
    case MHOD_COMMENT:
    case MHOD_CATEGORY:
    case MHOD_COMPOSER:
    case MHOD_GROUPING:
    case MHOD_DESCRIPTION_TEXT:
    case MHOD_SUBTITLE:
    // TODO check for these with the libqpod impl
    // case MHOD_TVSHOW:
    // case MHOD_TVEPISODE:
    // case MHOD_TVNETWORK:
    case MHOD_ALBUMARTIST:
    case MHOD_ARTIST_FOR_SORT:
    case MHOD_KEYWORDS:
    case MHOD_TITLE_FOR_SORT:
    case MHOD_ALBUM_FOR_SORT:
    case MHOD_ALBUMARTIST_FOR_SORT:
    case MHOD_COMPOSER_FOR_SORT:
        {
        uint ucs2len = ( blocklen - 40 ) / sizeof( Q_UINT16 );
        Q_UINT16 * buffer = new Q_UINT16 [ ucs2len + 1 ];

        itunesdb::utils::seekRelative( stream, 24 );    // skip stuff
        for ( uint i = 0; i < ucs2len; ++i ) {
            stream >> buffer[ i ];
        }

        buffer[ ucs2len ] = 0;
        setItemProperty( QString::fromUcs2( buffer ), ( ItemProperty ) type );

        delete [] buffer;
        }
        break;
    case MHOD_PODCAST_ENCLOSURE_URL:
    case MHOD_PODCAST_RSS_URL: {
        uint numBytes = blocklen - 24;
        char * buffer = new char [ numBytes + 1 ];

        itunesdb::utils::seekRelative( stream, 8 );
        stream.readRawBytes( buffer, numBytes );

        buffer[ numBytes ] = 0;
        setItemProperty( QString::fromUtf8( buffer ), ( ItemProperty ) type );

        delete [] buffer;
        }
        break;
    case MHOD_CHAPTER_DATA:
    case MHOD_SMART_PLAYLIST_INFO:
    case MHOD_SMART_PLAYLIST_IRULES:
    case MHOD_PLAYLIST:
    case MHOD_LIBRARY_PLAYLIST_INDEX:
        readNonStringMHOD( stream, type, blocklen );
        break;
    default:
        itunesdb::utils::seekRelative( stream, blocklen - 16 );  // skip the whole block
        break;    // ignore
    }
}

inline void writeUTF16StringTypeMHOD( QDataStream& stream, Q_UINT32 type, QString data ) {
    if ( data.isEmpty() ) {
        return;
    }

    const char *buf= (const char *)data.ucs2();
    if( buf == NULL) {
        return;
    }

    uint datalen= 2 * data.length();

    stream << (Q_UINT32) 0x646F686D;    // mhod
    stream << (Q_UINT32) 0x18;    // headerlen
    stream << (Q_UINT32) 40 + datalen;
    stream << (Q_UINT32) type;
    stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 1;    // dummy - would refer to the tracks position if used in MHOD_PLAYLIST
    stream << (Q_UINT32) datalen;
    stream << (Q_UINT32) 1;
    stream << (Q_UINT32) 0;
    stream.writeRawBytes( buf, datalen);
}

inline void writePodcastURLMHOD( QDataStream& stream, Q_UINT32 type, QString data ) {
    const char *buf= (const char *)data.ascii();
    if( buf == NULL) {
        return;
    }

    stream << (Q_UINT32) 0x646F686D;    // mhod
    stream << (Q_UINT32) 0x18;    // headerlen
    stream << (Q_UINT32) 24 + data.length();
    stream << (Q_UINT32) type;
    stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0;
    stream.writeRawBytes( data, data.length() );
}

inline void writeDumpedMHODToStream( QDataStream& stream, Q_UINT32 type, const QByteArray& data ) {
    if ( data.size() == 0 ) {
        return;
    }

    stream << (Q_UINT32) 0x646F686D;    // mhod
    stream << (Q_UINT32) 0x18;    // headerlen
    stream << (Q_UINT32) 16 + data.size();
    stream << (Q_UINT32) type;
    stream.writeRawBytes( data.data(), data.size() );
}

uint ListItem::writeNonStringMHODs( QDataStream& outstream ) const {
    uint num_mhods = 0;
    DumpedMHODs::const_iterator dumpedElement = m_dumpedMHODs.begin();
    for( ; dumpedElement != m_dumpedMHODs.end(); ++dumpedElement) {
        Q_UINT32 type = (*dumpedElement).first;
        switch ( type ) {
        case MHOD_CHAPTER_DATA:
        case MHOD_SMART_PLAYLIST_INFO:
        case MHOD_SMART_PLAYLIST_IRULES:
        case MHOD_PLAYLIST:
            writeDumpedMHODToStream( outstream, type, (*dumpedElement).second );
            num_mhods++;
            break;
        case MHOD_LIBRARY_PLAYLIST_INDEX:
            if ( !isDirty() ) { // this must be the library list
                writeDumpedMHODToStream( outstream, type, (*dumpedElement).second );
                num_mhods++;
            }
            break;
        default:
            break;    // ignore
        }

    }
    return num_mhods;
}

Q_UINT32 ListItem::writeMHODsToStream ( QDataStream & outstream ) const {
    Q_UINT32 num_mhods = 0;
    PropertyMap::const_iterator element = m_properties.begin();
    for( ; element != m_properties.end(); ++element) {
        if( (*element).second.isEmpty()) {
            continue;
        }

        Q_UINT32 type = (*element).first;
        switch ( type ) {
        case MHOD_TITLE:
        case MHOD_PATH:
        case MHOD_ALBUM:
        case MHOD_ARTIST:
        case MHOD_GENRE:
        case MHOD_FDESC:
        case MHOD_EQ_SETTING:
        case MHOD_COMMENT:
        case MHOD_CATEGORY:
        case MHOD_COMPOSER:
        case MHOD_GROUPING:
        case MHOD_DESCRIPTION_TEXT:
        case MHOD_SUBTITLE:
        // TODO check for these with the libqpod impl
        // case MHOD_TVSHOW:
        // case MHOD_TVEPISODE:
        // case MHOD_TVNETWORK:
        case MHOD_ALBUMARTIST:
        case MHOD_ARTIST_FOR_SORT:
        case MHOD_KEYWORDS:
        case MHOD_TITLE_FOR_SORT:
        case MHOD_ALBUM_FOR_SORT:
        case MHOD_ALBUMARTIST_FOR_SORT:
        case MHOD_COMPOSER_FOR_SORT:
            writeUTF16StringTypeMHOD( outstream, type, (*element).second );
            num_mhods++;
            break;
        case MHOD_PODCAST_ENCLOSURE_URL:
        case MHOD_PODCAST_RSS_URL:
            writePodcastURLMHOD( outstream, type, (*element).second );
            num_mhods++;
            break;
        default:
            break;    // ignore
        }

    }

    num_mhods += writeNonStringMHODs( outstream );

    return num_mhods;
}

void ListItem::doneAddingData() {
    // This method may be overridden for consistency checks after all properties have been set.
    // Does nothing here.
}

}
