 /*****************************************************************************
 *   Copyright (C) 2006 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include <qbuffer.h>

#include "playlist.h"
#include "utils.h"

#if QT_VERSION >= 0x040000
#define ptrWhenQT4(a) (&a)
#else
#define ptrWhenQT4(a) a
#endif


namespace itunesdb {


/**
 * The default playlist comparator comparing two given PlaylistItems by position.
 */
class DefaultPlaylistItemComparator : public PlaylistItemComparator {
public:
    virtual int compare( const PlaylistItem& item1, const PlaylistItem& item2 ) const {
        return item1.getPosition() - item2.getPosition();
    }
};



Playlist::Playlist()
    : ListItem( ITEMTYPE_PLAYLIST ), m_tracklist( true ), m_isMaster( 0 ), m_flag2( 0 ),
      m_flag3( 0 ), m_flag4( 0 ), timeStamp( 0 ), m_id( 0 ), unk3( 0 ), mIsPodcast( 0 ),
      m_order( SORTORDER_MANUAL ), mSplRuleSet( NULL )
{
    m_tracklist.setComparator( new DefaultPlaylistItemComparator() );
}


Playlist::Playlist( const Playlist& src )
    : ListItem( src ), m_tracklist( true ), m_isMaster( src.m_isMaster ),
      timeStamp( src.timeStamp ), m_id( src.m_id ), unk3( src.unk3 ),
      mIsPodcast( src.mIsPodcast ), m_order( src.m_order ), mSplRuleSet( NULL )
{
    // TODO ATTENTION smart playlist rules won't get copied
    // TODO ATTENTION tracks won't get copied
    // printf( "copying playlist %s\n", src.getTitle().ascii() );
    m_tracklist.setComparator( src.m_tracklist.getComparator() );
}


Playlist::~Playlist()
{
    // printf( "playlist %s\n", getTitle().ascii() );
    m_tracklist.clear();
    delete mSplRuleSet;
}

QDataStream & Playlist::readFromStream( QDataStream& instream, bool * ok ) {
    Q_UINT32 dummy, blocklen;
    instream >> blocklen;

    // 8 bytes read

    if ( blocklen < 48 ) {
        // insufficient information
        QByteArray buffer( blocklen );
        instream.readRawBytes(buffer.data(), blocklen - 8);

        if ( ok ) {
            *ok = false;
        }

        return instream;
    }

    instream >> dummy;  // length including all child records
    instream >> dummy;  // number of data objects (mhod)
    instream >> dummy;  // number of tracks

    // 20 bytes read

    instream >> m_isMaster;
    instream >> m_flag2;
    instream >> m_flag3;
    instream >> m_flag4;
    instream >> timeStamp;
    instream >> m_id;

    // 36 bytes read

    instream >> unk3;  // unk3

    instream >> mIsPodcast;  // number of string based MHODs    - ignore
    instream >> mIsPodcast;  // isPodcast flag

    instream >> dummy; m_order = ( Sortorder )dummy;

    int rest = blocklen - 48;
    if ( rest > 0 ) {
        itunesdb::utils::seekRelative( instream, rest );
    }
    setDirty( false );

    if ( ok ) {
        *ok = true;
    }

    return instream;
}


const QString& Playlist::getTitle() const
{
    return getItemProperty( MHOD_TITLE);
}

Playlist::Sortorder Playlist::getSortOrder() const {
    return m_order;
}


void Playlist::doneAddingData() {
    if( getItemProperty( MHOD_TITLE ).isEmpty() ) {
        setItemProperty("_no_title_", MHOD_TITLE);
    }
}

/*!
    \fn Playlist::setTitle( QString& newtitle)
 */
void Playlist::setTitle( const QString& newtitle )
{
    setItemProperty( newtitle, MHOD_TITLE );
    doneAddingData();    // consisteny check
    setDirty();
}


bool Playlist::contains( Q_UINT32 trackID ) const {
    ConstIterator trackiter = getElements();
    while( trackiter.hasNext() ) {
        if( trackiter.next()->getID() == trackID ) {
            return true;
        }
    }
    return false;
}


PlaylistItem * Playlist::createNewItem( Q_UINT32 itemid ) {
    return new PlaylistItem( itemid );
}


int Playlist::addPlaylistItem( Q_UINT32 trackid, int position ) {
    PlaylistItem * item = createNewItem( trackid );

    if ( position > 0 && (uint)position < m_tracklist.count() ) {
        item->setPosition( position );
    }

    int result = addPlaylistItem( item );
    if ( result >= 0 ) {
        setDirty();
    }

    return result;
}


int Playlist::addPlaylistItem( PlaylistItem * newItem ) {
    if ( !newItem ) {
        return -1;
    }

    bool setPos = !newItem->isPositionSet();
    if ( setPos ) { // position isn't set
        // set it to a sane value, so inSort() will work with the default comparator
        if ( m_order == SORTORDER_RANDOM ) {
            uint trackcount = m_tracklist.count();
            newItem->setPosition( ( trackcount > 0 ? rand() % trackcount : 0 ) + 1 );
        } else {
            newItem->setPosition( m_tracklist.count() + 1 );
        }
    }

    // add it at the correct position
    m_tracklist.inSort( newItem );

    if ( setPos ) {
        updatePositions();
    }

    return newItem->getPosition() - 1;
}


struct TrackIDPredicate {
    Q_UINT32 m_trackid;
    TrackIDPredicate( Q_UINT32 trackid )
        : m_trackid( trackid ) {}
    bool operator() ( const itunesdb::PlaylistItem * item ) const {
        return item && item->getID() == m_trackid;
    }
};
bool Playlist::removeAll( Q_UINT32 trackid ) {
    return m_tracklist.filteredElements( TrackIDPredicate( trackid ) ).removeRemaining();
}


Q_UINT32 Playlist::removeTrackAt( Iterator& pos ) {
    const PlaylistItem * item = pos.current();
    Q_UINT32 result = item ? item->getID() : PLAYLISTITEM_INVALID;

    pos.remove();

    setDirty();

    return result;
}


Q_UINT32 Playlist::removeTrackAt( uint pos ) {
    const PlaylistItem * item = m_tracklist[ pos ];
    Q_UINT32 result = item ? item->getID() : PLAYLISTITEM_INVALID;
    if ( !m_tracklist.remove( pos ) ) {
        return PLAYLISTITEM_INVALID;
    }

    return result;
}


Q_UINT32 Playlist::getTrackIDAt( uint pos ) {
    PlaylistItem* item;
    if( pos < m_tracklist.count() && ( item = m_tracklist[ pos ] ) ) {
        return item->getID();
    } else {
        return PLAYLISTITEM_INVALID;
    }
}


// FIXME container might get changed here - we need to handle the "dirty" flag
Playlist::Iterator Playlist::getElements() {
    return m_tracklist.iterator();
}


Playlist::ConstIterator Playlist::getElements() const {
    return m_tracklist.iterator();
}


/*!
    \fn itunesdb::Playlist::getNumTracks()
 */
uint Playlist::getNumTracks() const
{
    return m_tracklist.count();
}


void Playlist::setID( Q_UINT64 id ) {
    m_id = id;
}


void Playlist::clear() {
    m_tracklist.clear();
    setDirty();
}


void Playlist::sort() {
    if ( getSortOrder() == SORTORDER_RANDOM ) {
        m_tracklist.randomize();
    } else {
        m_tracklist.sort();
    }

    updatePositions();
}


SmartPlaylistRuleSet& Playlist::enableSmartPlaylist() {
    if ( mSplRuleSet == NULL ) {
        mSplRuleSet = createNewSplRuleSet();
    }
    return *mSplRuleSet;
}


bool Playlist::isSmartPlaylist() const {
    return mSplRuleSet != NULL;
}


void Playlist::discardSmartPlaylistData() {
    SmartPlaylistRuleSet * ptr = mSplRuleSet;
    mSplRuleSet = NULL;
    delete ptr;
}


SmartPlaylistRuleSet * Playlist::getSmartPlaylistRules() const {
    return mSplRuleSet;
}


SmartPlaylistRuleSet * Playlist::createNewSplRuleSet() {
    return new SmartPlaylistRuleSet();
}


void Playlist::readNonStringMHOD( QDataStream& stream, Q_UINT32 type, Q_UINT32 blocklen ) {
    switch ( type ) {
    case MHOD_SMART_PLAYLIST_INFO: {
            SmartPlaylistRuleSet& ruleSet = enableSmartPlaylist();
            Q_UINT32 dummy; stream >> dummy; stream >> dummy;
            ruleSet.readHeader( stream, blocklen - 24 );
        }
        break;
    case MHOD_SMART_PLAYLIST_IRULES: {
            SmartPlaylistRuleSet& ruleSet = enableSmartPlaylist();
            Q_UINT32 dummy; stream >> dummy; stream >> dummy;
            stream.setByteOrder( QDataStream::BigEndian );
            ruleSet.readRules( stream, blocklen - 24 );
            stream.setByteOrder( QDataStream::LittleEndian );
        }
        break;
    default:
        ListItem::readNonStringMHOD( stream, type, blocklen );
    }
}

void Playlist::fillSplHeaderBuffer( QByteArray& buffer ) const {
    QDataStream stream ( ptrWhenQT4(buffer), IO_WriteOnly );
    stream.setByteOrder( QDataStream::LittleEndian);

    if ( mSplRuleSet ) {
        mSplRuleSet->writeHeader( stream );
    }
}


void Playlist::fillSplRulesBuffer( QByteArray& buffer ) const {
    QDataStream stream ( ptrWhenQT4(buffer), IO_WriteOnly );
    stream.setByteOrder( QDataStream::BigEndian );

    if ( mSplRuleSet ) {
        mSplRuleSet->writeRules( stream );
    }
}


uint Playlist::writeNonStringMHODs( QDataStream& outstream ) const {
    uint result = ListItem::writeNonStringMHODs( outstream );

    if ( isSmartPlaylist() ) {
        QByteArray headerBuffer;
        fillSplHeaderBuffer( headerBuffer );

        outstream << (Q_UINT32) 0x646F686D;    // mhod
        outstream << (Q_UINT32) 0x18;    // headerlen
        outstream << (Q_UINT32) 0x18 + headerBuffer.size();
        outstream << (Q_UINT32) MHOD_SMART_PLAYLIST_INFO;
        outstream << (Q_UINT32) 0;
        outstream << (Q_UINT32) 0;

        outstream.writeRawBytes( headerBuffer.data(), headerBuffer.size());
        result++;

        QByteArray rulesBuffer;
        fillSplRulesBuffer( rulesBuffer );

        outstream << (Q_UINT32) 0x646F686D;    // mhod
        outstream << (Q_UINT32) 0x18;    // headerlen
        outstream << (Q_UINT32) 0x18 + rulesBuffer.size();
        outstream << (Q_UINT32) MHOD_SMART_PLAYLIST_IRULES;
        outstream << (Q_UINT32) 0;
        outstream << (Q_UINT32) 0;

        outstream.writeRawBytes( rulesBuffer.data(), rulesBuffer.size());
        result++;

    }

    return result;
}


/*!
    \fn itunesdb::Playlist::writePlaylistData( QByteArray& data)
 */
void Playlist::writeTitle( QDataStream& stream ) const
{
    QString title = getTitle();
    if ( title.isEmpty() ) {
        title = "KPod";
    }
    const char *data = (const char *)title.ucs2();
    if( data == NULL ) {
        return;
    }

    int datalen= 2* title.length();
    stream << (Q_UINT32) 0x646F686D;    // mhod
    stream << (Q_UINT32) 0x18;
    stream << (Q_UINT32) 40+ datalen;
    stream << (Q_UINT32) MHOD_TITLE;
    stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 1;    // dummy - would refer to the trackID if used in playlist
    stream << (Q_UINT32) datalen;
    stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0;
    stream.writeRawBytes( data, datalen);
}

/*!
    writes the stuff uiTunes would write as the second mhod for playlists. We don't use it - it's just here for completeness
 */
void Playlist::writeLongPlaylist( QDataStream& stream ) const
{
    stream << (Q_UINT32) 0x646F686D;    // mhod
    stream << (Q_UINT32) 0x18;          // size of header
    stream << (Q_UINT32) 0x0288;        // size of header + body
    stream << (Q_UINT32) MHOD_PLAYLIST; // type of the entry
    for( int i= 0; i< 6; i++)
        stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0x010084;      // unknown
    stream << (Q_UINT32) 0x05;              // unknown
    stream << (Q_UINT32) 0x09;              // unknown
    stream << (Q_UINT32) 0x03;              // unknown
    stream << (Q_UINT32) 0x120001;          // unknown
    for( int i= 0; i< 3; i++)
        stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0xc80002;          // unknown
    for( int i= 0; i< 3; i++)
        stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0x3c000d;          // unknown
    for( int i= 0; i< 3; i++)
        stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0x7d0004;          // unknown
    for( int i= 0; i< 3; i++)
        stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0x7d0003;          // unknown
    for( int i= 0; i< 3; i++)
        stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0x640008;          // unknown
    for( int i= 0; i< 3; i++)
        stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0x640017;          // unknown
    stream << (Q_UINT32) 0x01;              // bool? (visible? / colums?)
    for( int i= 0; i< 2; i++)
        stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0x500014;          // unknown
    stream << (Q_UINT32) 0x01;              // bool? (visible?)
    for( int i= 0; i< 2; i++)
        stream << (Q_UINT32) 0;
    stream << (Q_UINT32) 0x7d0015;          // unknown
    stream << (Q_UINT32) 0x01;              // bool? (visible?)
    for( int i= 0; i< 114; i++)
        stream << (Q_UINT32) 0;
}

/**
 * Writes the track list
 */
void Playlist::writeTracks( QDataStream& stream ) const
{
    // write playlist items
    ConstIterator trackiterator= getElements();
    int pos = 1;
    while( trackiterator.hasNext()) {
        Q_UINT32 trackid= trackiterator.next()->getID();

        stream << (Q_UINT32) 0x7069686D;    // mhip
        stream << (Q_UINT32) 0x4C;    // headerlen
        stream << (Q_UINT32) 0x78;    // datalen
        stream << (Q_UINT32) 1;
        stream << (Q_UINT32) 0;
        stream << (Q_UINT32) pos;
        stream << (Q_UINT32) trackid;
        for ( int i = 0; i < 12; i++) {
            stream << (Q_UINT32) 0;
        }
        stream << (Q_UINT32) 0x646F686D;   // mhod
        stream << (Q_UINT32) 0x18;
        stream << (Q_UINT32) 0x2C;
        stream << (Q_UINT32) MHOD_PLAYLIST;    // type: playlist
        stream << (Q_UINT32) 0;
        stream << (Q_UINT32) 0;
        stream << (Q_UINT32) pos;
        for( int i= 0; i< 4; i++)
            stream << (Q_UINT32) 0;
        pos++;
    }
}


/*!
    \fn itunesdb::Playlist::writePlaylistData( QByteArray& data)
 */
void Playlist::writeData( QByteArray& data, bool isMainlist) const
{
    QBuffer buffer( ptrWhenQT4(data) );

    buffer.open(IO_WriteOnly);
    QDataStream stream( &buffer);
    stream.setByteOrder( QDataStream::LittleEndian);

    // write the header
    stream << (Q_UINT32) 0x7079686D;    // mhyp
    stream << (Q_UINT32) 0x6C;    // headerlen
    stream << (Q_UINT32) 0x0;    // length - set later
    stream << (Q_UINT32) getNumComponents();
    stream << (Q_UINT32) getNumTracks();
    stream << (Q_UINT8)  (isMainlist ? 1 : 0);  // 1=hidden (library lis), 2=visible
    stream << (Q_UINT8)  m_flag2;
    stream << (Q_UINT8)  m_flag3;
    stream << (Q_UINT8)  m_flag4;
    stream << (Q_UINT32) timeStamp;
    stream << m_id;
    stream << (Q_UINT32) unk3;     // unknown, but significant
    stream << (Q_UINT16) getNumStringComponents();     // number of string MHODs
    stream << (Q_UINT16) mIsPodcast;     // normal PL:0, podcast PL:1
    stream << (Q_UINT32) (isMainlist ? 8 : m_order);  // sort order

    for( int i= 0; i< 15; i++) {
        stream << (Q_UINT32) 0;    // pad the rest
    }

    uint numMHODs = writeMHODsToStream( stream );

    writeTracks( stream );

    buffer.at( 8);
    stream << (Q_UINT32) data.size();
    stream << (Q_UINT32) numMHODs;
    buffer.close();
}

/**
*/
QDataStream & Playlist::writeToStream(QDataStream & outstream, bool isMainlist) {
    QByteArray buffer;

    m_tracklist.sort();

    writeData(buffer, isMainlist);
    outstream.writeRawBytes( buffer.data(), buffer.size());

    return outstream;
}


void Playlist::updatePositions() {
    // normalize the playlist
    uint pos = 0;
    TrackList_T::Iterator iter = m_tracklist.iterator();
    while ( iter.hasNext() ) {
        PlaylistItem * current = iter.next();
        if ( !current ) {
            continue;
        }
        current->setPosition( ++pos );
    }
}


}
