 /*****************************************************************************
 *   Copyright (C) 2006 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include "smartplaylistrules.h"
#include "utils.h"

#include <qdatastream.h>
#include <qdatetime.h>

namespace itunesdb {


SmartPlaylistRule::SmartPlaylistRule( SPLStringField field, SPLStringAction action,
                                      const QString& testString )
    : mField( field), mAction( action ), mString( testString ),
      mFromValue( 0 ), mToValue( 0 ),
      mFromDate( 0 ), mToDate( 0 ),
      mFromUnits( 0 ), mToUnits( 0 ),
      mCalculatedFrom( 0 ), mCalculatedTo( 0 )
{
}


SmartPlaylistRule::SmartPlaylistRule( SPLUIntField field, SPLUIntAction action,
                                      Q_UINT64 fromValue, Q_UINT64 toValue,
                                      Q_INT64  fromDate,  Q_INT64  toDate,
                                      Q_UINT64 fromUnits, Q_UINT64 toUnits )
    : mField( field), mAction( action ),
      mFromValue( fromValue ), mToValue( toValue ),
      mFromDate( fromDate ), mToDate( toDate ),
      mFromUnits( fromUnits ), mToUnits( toUnits ),
      mCalculatedFrom( calculateTestValue( fromValue, fromDate, fromUnits ) ),
      mCalculatedTo( calculateTestValue( toValue, toDate, toUnits ) )
{
}


SmartPlaylistRule::~SmartPlaylistRule() {
}


Q_UINT64 SmartPlaylistRule::calculatedFrom() const {
    if ( mFromValue == SPL_VALUE_NOW ) {
        return calculateTestValue( mFromValue, mFromDate, mFromUnits );
    }
    return mCalculatedFrom;
}


Q_UINT64 SmartPlaylistRule::calculatedTo() const {
    if ( mToValue == SPL_VALUE_NOW ) {
        return calculateTestValue( mToValue, mToDate, mToUnits );
    }
    return mCalculatedTo;
}


inline void calculateMacDate( Q_UINT64& unixDate ) {
    if ( unixDate && unixDate != SPL_VALUE_NOW ) {
        unixDate += MAC_EPOCH_DELTA;
    }
}

uint SmartPlaylistRule::writeToStream( QDataStream& outStream ) const {
    uint result = 52;

    outStream << (Q_UINT32) getField();
    outStream << (Q_UINT32) getAction();

    // 44 bytes of padding
    for ( uint i = 0; i < 11; i++ ) {
        outStream << (Q_UINT32) 0;
    }

    if ( isStringBasedRule() ) {
        const char *buf= (const char *)getString().ucs2();
        Q_UINT32 numBytes = 2 * getString().length();

        outStream << numBytes;
        if ( numBytes && buf ) {
            outStream.writeRawBytes( buf, numBytes );
        }

        result += 4 + numBytes;

    } else {
        Q_UINT64 fromValue = getFromValue();
        Q_UINT64 toValue = getToValue();
        Q_INT64 fromDate = getFromDate();
        Q_INT64 toDate = getToDate();

        if ( fromDate || toDate ) {
            calculateMacDate( fromValue );
            calculateMacDate( toValue );
        }

        outStream << (Q_UINT32) 0x44;
        outStream << fromValue;
        outStream << fromDate;
        outStream << (Q_UINT64) getFromUnits();

        outStream << toValue;
        outStream << toDate;
        outStream << (Q_UINT64) getToUnits();

        // 20 bytes of padding
        for ( uint i = 0; i < 5; i++ ) {
            outStream << (Q_UINT32) 0;
        }

        result += 0x48;

    }

    return result;
}


const QString SmartPlaylistRule::toString() const {
    QString result( getFieldString( mField ) + " " + getActionString( mAction ) + " " );

    // create textual representation for the two kinds of rules.
    if ( isStringBasedRule() ) {
        return  result + mString;

    } else {
        return result +
               QString::number( calculatedFrom() ) + "/" +
               QString::number( calculatedTo() );

    }
}


QString SmartPlaylistRule::getActionString( uint action ) {
    switch( action ) {
    case ACTION_IS_STRING:
    case ACTION_IS_UINT:
        return QString( "is" );
    case ACTION_CONTAINS:
        return QString( "contains" );
    case ACTION_STARTSWITH:
        return QString( "starts with" );
    case ACTION_ENDSWITH:
        return QString( "ends with" );
    case ACTION_IS_NOT_STRING:
    case ACTION_IS_NOT_UINT:
        return QString( "is not" );
    case ACTION_DOESNT_CONTAIN:
        return QString( "does not contain" );
    case ACTION_DOESNT_STARTWITH:
        return QString( "does not start with" );
    case ACTION_DOESNT_ENDWITH:
        return QString( "does not ent with" );
    case ACTION_GREATER_THAN:
        return QString( "is greater than" );
    case ACTION_GREATER_OR_EQUAL:
        return QString( "is greater or equal" );
    case ACTION_LESS_THAN:
        return QString( "is less than" );
    case ACTION_LESS_OR_EQUAL:
        return QString( "is less or equal" );
    case ACTION_IN_RANGE:
        return QString( "is in the range" );
    case ACTION_IN_THE_LAST:
        return QString( "is in the last" );
    case ACTION_NOT_GREATER:
        return QString( "is not greater" );
    case ACTION_NOT_GREATER_OR_EQUAL:
        return QString( "is greater or equal" );
    case ACTION_NOT_LESS:
        return QString( "is not less than" );
    case ACTION_NOT_LESS_OR_EQUAL:
        return QString( "is not less or equal" );
    case ACTION_NOT_IN_RANGE:
        return QString( "is not in the range" );
    case ACTION_NOT_IN_THE_LAST:
        return QString( "is not in the last" );
    }
    return itunesdb::utils::NullQString;
}


QString SmartPlaylistRule::getFieldString( uint field ) {
    switch ( field ) {
    case FIELD_NAME:
        return QString( "title" );
    case FIELD_ALBUM:
        return QString( "album" );
    case FIELD_ARTIST:
        return QString( "artist" );
    case FIELD_GENRE:
        return QString( "genre" );
    case FIELD_KIND:
        return QString( "kind" );
    case FIELD_COMMENT:
        return QString( "comment" );
    case FIELD_COMPOSER:
        return QString( "composer" );
    case FIELD_GROUPING:
        return QString( "grouping" );
    case FIELD_DESCRIPTION:
        return QString( "description" );
    case FIELD_CATEGORY:
        return QString( "category" );
    case FIELD_BITRATE:
        return QString( "bitrate" );
    case FIELD_SAMPLERATE:
        return QString( "samplerate" );
    case FIELD_YEAR:
        return QString( "year" );
    case FIELD_DATEMODIFIED:
        return QString( "modification time" );
    case FIELD_TRACKNUMBER:
        return QString( "track number" );
    case FIELD_SIZE:
        return QString( "size" );
    case FIELD_TIME:
        return QString( "length" );
    case FIELD_DATEADDED:
        return QString( "date added" );
    case FIELD_PLAYCOUNT:
        return QString( "playcount" );
    case FIELD_LASTPLAYED:
        return QString( "last played" );
    case FIELD_DISCNUMBER:
        return QString( "disc number" );
    case FIELD_RATING:
        return QString( "rating" );
    case FIELD_COMPILATION:
        return QString( "compilation" );
    case FIELD_BPM:
        return QString( "bpm" );
    case FIELD_PLAYLISTID:
        return QString( "playlist ID" );
    case FIELD_PODCAST:
        return QString( "podcast" );
    }
    return itunesdb::utils::NullQString;
}


Q_UINT64 SmartPlaylistRule::calculateTestValue( Q_UINT64 value, Q_INT64 date, Q_UINT64 units ) {
    Q_INT64 result;
    if ( value == SPL_VALUE_NOW ) {
        result = QDateTime::currentDateTime().toTime_t();
    } else {
        result = value;
    }
    return result + date * units;
}


/***
 *
 * SmartPlaylistRuleSet
 *
 ***/


SmartPlaylistRuleSet::SmartPlaylistRuleSet()
    : mMatchAny( false ), unk5( 0x00010001 ), mLiveUpdate( 0 ),
      mCheckLimits( 0 ), mCheckRules( 0 ), mCheckedOnly( 0 ),
      mReverseLimitSort( 0 ), mLimitSort( LIMIT_SORT_ARTIST ),
      mLimitType( itunesdb::LIMIT_TYPE_SONGS ), mLimitValue( 0 )
{
}


SmartPlaylistRuleSet::SmartPlaylistRuleSet( const SmartPlaylistRuleSet & src )
    : mMatchAny( src.mMatchAny ), unk5( src.unk5 ),
      mLiveUpdate( src.mLiveUpdate ), mCheckLimits( src.mCheckLimits ),
      mCheckRules( src.mCheckRules ), mCheckedOnly( src.mCheckedOnly ),
      mReverseLimitSort( src.mReverseLimitSort ),
      mLimitSort( src.mLimitSort ), mLimitType( src.mLimitType ),
      mLimitValue( src.mLimitValue )
{
}


SmartPlaylistRuleSet::~SmartPlaylistRuleSet()
{
    clear();
}


bool SmartPlaylistRuleSet::removeRuleAt( uint position ) {
    if ( position > mRules.size() ) {
        return false;
    }

    // position an iterator at the element to be removed
    RuleList::iterator ruleIter = mRules.begin() + position;
    SmartPlaylistRule * rule = *ruleIter;
    mRules.erase( ruleIter );
    delete rule;

    return true;
}


void SmartPlaylistRuleSet::clear() {
    // delete the rules
    RuleList::iterator ruleIter = mRules.begin();
    for( ; ruleIter != mRules.end(); ++ruleIter ) {
        SmartPlaylistRule * rule = *ruleIter;
        delete rule;
    }
    mRules.clear();
}


bool SmartPlaylistRuleSet::addStringRule(
    SPLStringField field, SPLStringAction action, const QString& testString, int position )
{
    SmartPlaylistRule * rule = createStringRule( field, action, testString );
    if ( rule ) {
        if ( position < 0 || (uint)position >= mRules.size() ) {
            mRules.push_back( rule );
        } else {
            mRules.insert( mRules.begin() + position, rule );
        }
        mCheckRules = true;
        return true;
    }

    return false;
}


SmartPlaylistRule * SmartPlaylistRuleSet::createStringRule( SPLStringField field,
    SPLStringAction action, const QString& testString )
{
    return new SmartPlaylistRule( field, action, testString );
}


bool SmartPlaylistRuleSet::addUIntRule(
                                SPLUIntField field, SPLUIntAction action,
                                Q_UINT64 fromValue, Q_UINT64 toValue,
                                 Q_INT64 fromDate,   Q_INT64 toDate,
                                Q_UINT64 fromUnits, Q_UINT64 toUnits,
                                int position )
{
    SmartPlaylistRule * rule =
        createUIntRule( field, action, fromValue, toValue,
                        fromDate, toDate, fromUnits, toUnits );
    if ( rule ) {
        if ( position < 0 || (uint)position >= mRules.size() ) {
            mRules.push_back( rule );
        } else {
            mRules.insert( mRules.begin() + position, rule );
        }
        mCheckRules = true;
        return true;
    }

    return false;
}


SmartPlaylistRule * SmartPlaylistRuleSet::createUIntRule(  SPLUIntField field, SPLUIntAction action,
                                     Q_UINT64 fromValue, Q_UINT64 toValue,
                                     Q_INT64  fromDate,  Q_INT64  toDate,
                                     Q_UINT64 fromUnits, Q_UINT64 toUnits )
{
    return new SmartPlaylistRule( field, action, fromValue, toValue, fromDate, toDate, fromUnits, toUnits );
}



inline void calculateUnixDate( Q_UINT64& macDate ) {
    if ( macDate && macDate != SPL_VALUE_NOW ) {
        macDate -= MAC_EPOCH_DELTA;
    }
}
SmartPlaylistRule * SmartPlaylistRuleSet::readSingleRule( QDataStream & instream, uint * bytesRead ) {
    SmartPlaylistRule * rule = NULL;

    Q_UINT32 field; instream >> field;
    Q_UINT32 action; instream >> action;

    // read 44 bytes of padding
    itunesdb::utils::seekRelative( instream, 44 );

    Q_UINT32 length; instream >> length;

    if ( action & ACTION_STRING_BIT ) {
        Q_UINT16 * buffer = new Q_UINT16 [ length + 1 ];

        for ( uint i = 0; i < length; ++i ) {
            instream >> buffer[ i ];
        }
        buffer[ length ] = 0;

        rule = createStringRule( (SPLStringField)field, (SPLStringAction)action, QString::fromUcs2( buffer ) );

        delete [] buffer;

    } else {

        Q_UINT64 fromValue; instream >> fromValue;
        Q_UINT64 fromDate; instream >> fromDate;
        Q_UINT64 fromUnits; instream >> fromUnits;

        Q_UINT64 toValue; instream >> toValue;
        Q_UINT64 toDate; instream >> toDate;
        Q_UINT64 toUnits; instream >> toUnits;

        if ( fromDate || toDate ) {
            calculateUnixDate( fromValue );
            calculateUnixDate( toValue );
        }

        itunesdb::utils::seekRelative( instream, length - 48 );    // read the rest

        rule = createUIntRule( (SPLUIntField)field, (SPLUIntAction)action,
                               fromValue, toValue,
                               fromDate, toDate,
                               fromUnits, toUnits );

    }

    if ( bytesRead != NULL ) {
        *bytesRead = length + 56;
    }

    return rule;
}


uint SmartPlaylistRuleSet::readRules( QDataStream & instream, uint length ) {
    Q_UINT32 dummy, numRules;

    instream >> dummy;  // SLst
    instream >> unk5;

    instream >> numRules;
    instream >> dummy; mMatchAny = dummy;

    itunesdb::utils::seekRelative( instream, 120 );

    uint bytesleft = length - 136;

    for ( uint ruleIdx = 0;
          ruleIdx < numRules && bytesleft >= 56 && !instream.atEnd();
          ruleIdx++ )
    {
        uint numRead = 0;
        SmartPlaylistRule * rule = readSingleRule( instream, &numRead );
        if ( rule ) {
            mRules.push_back( rule );
        }
        bytesleft -= numRead;
    }

    if ( bytesleft > 0 && !instream.atEnd() ) {
        itunesdb::utils::seekRelative( instream, bytesleft );
    }

    return numRules;
}


bool SmartPlaylistRuleSet::readHeader( QDataStream& instream, uint length ) {

    if ( length < 14 ) {
        return false;
    }

    Q_UINT8 rdByte = 0;

    instream >> mLiveUpdate;
    instream >> mCheckRules;
    instream >> mCheckLimits;
    instream >> rdByte; mLimitType = ( SPLLimitType )rdByte;
    instream >> rdByte; mLimitSort = ( SPLLimitSort )rdByte;
    instream >> rdByte;
    instream >> rdByte;
    instream >> rdByte;
    instream >> mLimitValue;
    instream >> mCheckedOnly;
    instream >> mReverseLimitSort;

    itunesdb::utils::seekRelative( instream, length - 14 );

    return true;
}


uint SmartPlaylistRuleSet::writeRules( QDataStream & outstream ) const {
    uint result = 136;

    outstream << (Q_UINT32) 0x534C7374;  // SLst
    outstream << (Q_UINT32) unk5;  // unk5
    outstream << (Q_UINT32) mRules.size();
    outstream << ( (Q_UINT32) mMatchAny ? 1 : 0 );

    for ( uint i = 0; i < 30; i++ ) {
        outstream << (Q_UINT32) 0;
    }

    RuleList::const_iterator ruleIter = mRules.begin();
    for ( ; ruleIter != mRules.end(); ++ruleIter ) {
        result += (*ruleIter)->writeToStream( outstream );
    }

    return result;
}


uint SmartPlaylistRuleSet::writeHeader( QDataStream& outstream ) const {
    outstream << mLiveUpdate;
    outstream << mCheckRules;
    outstream << mCheckLimits;
    outstream << (Q_UINT8) mLimitType;
    outstream << (Q_UINT8) mLimitSort;
    outstream << (Q_UINT8) 0;
    outstream << (Q_UINT8) 0;
    outstream << (Q_UINT8) 0;
    outstream << mLimitValue;
    outstream << mCheckedOnly;
    outstream << mReverseLimitSort;

    // pad the rest (58 bytes)
    outstream << (Q_UINT16) 0;
    for ( uint i = 0; i < 14; i++ ) {
        outstream << (Q_UINT32) 0;
    }

    return 72;
}


void SmartPlaylistRuleSet::setLiveUpdateFlag( bool liveupdate ) {
    mLiveUpdate = liveupdate;
}


void SmartPlaylistRuleSet::setCheckRulesFlag( bool checkRules ) {
    mCheckRules = checkRules;
}


void SmartPlaylistRuleSet::setLimits( SPLLimitSort limitSort,
                                      SPLLimitType limitType,
                                      Q_UINT32 limitValue,
                                      bool reverseLimitSort )
{
    mLimitSort = limitSort;
    mLimitType = limitType;
    mLimitValue = limitValue;
    mReverseLimitSort = reverseLimitSort ? 1 : 0;
    mCheckLimits = true;
}


void SmartPlaylistRuleSet::enableLimits( bool enable ) {
    mCheckLimits = enable;
}


void SmartPlaylistRuleSet::setMatchAnyFlag( bool matchAny ) {
    mMatchAny = matchAny;
}


void SmartPlaylistRuleSet::setCheckedTracksOnlyFlag( bool checkedOnly ) {
    mCheckedOnly = checkedOnly;
}


const itunesdb::RuleList& SmartPlaylistRuleSet::getRules() const {
    return mRules;
}


uint SmartPlaylistRuleSet::getNumRules() const {
    return mRules.size();
}


QString SmartPlaylistRuleSet::toString() const {
    QString result( "all tracks where " );
    if ( isMatchAny() ) {
        result.append( "any rule applies [\n" );
    } else {
        result.append( "all rules apply [\n" );
    }
    RuleList::const_iterator ruleIter = mRules.begin();
    for ( ; ruleIter != mRules.end(); ++ruleIter ) {
        result.append( "\t" );
        result.append( (*ruleIter)->toString() );
        result.append( "\n" );
    }
    result.append( "]\n" );
    return result;
}


}

